from __future__ import unicode_literals, print_function

from itertools import izip_longest

import pytest

from treeoflife.test.util import FakeNodeCreator, match
from treeoflife.tracker import Tracker
from treeoflife.nodes.node import Node, _NodeListRoot, Option
from treeoflife.nodes.misc import GenericNode, GenericActivate
from treeoflife.file_storage import serialize_to_str
from treeoflife import exceptions
from treeoflife import navigation


class TestNode(object):
    def test_iter_parents(self):
        tracker = Tracker(skeleton=False)

        input_str = (
            "_genactive: 0\n"
            "    _genactive: 1\n"
            "        _genactive: 2\n"
            "            _genactive: 3\n"
            "                @active"
        )

        tracker.deserialize("str", input_str)

        expected_pairs = [
            ("_genactive", "3"),
            ("_genactive", "2"),
            ("_genactive", "1"),
            ("_genactive", "0"),
            ("life", None)
        ]

        for expected, node in izip_longest(expected_pairs,
                tracker.root.active_node.iter_parents()):
            node_type, text = expected
            assert node_type == node.node_type
            assert text == node.text

    def tracker(self):
        return Tracker(nodecreator=FakeNodeCreator(Node),
                skeleton=False)

    def test_multiline_init(self):
        tracker = self.tracker()

        class SubNode(Node):
            multiline = False

        with pytest.raises(exceptions.LoadError):
            SubNode("subnode", "text\nothertext", tracker.root)

    def test_multiline_continue(self):
        tracker = self.tracker()

        class SubNode(Node):
            multiline = False

        subnode = SubNode("subnode", "text", tracker.root)
        with pytest.raises(exceptions.LoadError):
            subnode.continue_text("line 2")

    def test_toplevel(self):
        tracker = self.tracker()

        class SubNode(Node):
            toplevel = True

        middle = SubNode("middle", None, tracker.root)

        node = SubNode("error", None, middle)
        with pytest.raises(exceptions.LoadError):
            node._validate()

    def test_textless(self):
        tracker = self.tracker()

        class SubNode(Node):
            textless = True

        success = SubNode("subnode", None, tracker.root)
        assert success

        with pytest.raises(exceptions.LoadError):
            SubNode("subnode", "unallowed text", tracker.root)

    def test_notext_continueline(self):
        tracker = self.tracker()

        class SubNode(Node):
            multiline = True

        node = SubNode("node", None, tracker.root)

        with pytest.raises(exceptions.LoadError):
            node.continue_text("error")

    def test_allowed_children(self):
        tracker = self.tracker()

        class ParentNode(Node):
            allowed_children = ("allowed",)

        parent = ParentNode("parent", None, tracker.root)

        child1 = Node("allowed", None, parent)
        parent.addchild(child1)

        child2 = Node("not-allowed", None, parent)
        with pytest.raises(exceptions.LoadError):
            parent.addchild(child2)

    def test_allowed_parents(self):
        tracker = self.tracker()

        class ChildNode(Node):
            children_of = ("allowed-parent",)

        allowed_parent = Node("allowed-parent", None, tracker.root)
        child1 = ChildNode("child1", None, allowed_parent)
        child1._validate()
        allowed_parent.addchild(child1)

        disallowed_parent = Node("some-other-node", None, tracker.root)
        child2 = ChildNode("child2", None, disallowed_parent)
        with pytest.raises(exceptions.LoadError):
            child2._validate()

    def test_remove_child(self):
        tracker = self.tracker()

        parent1 = Node("parent1", None, tracker.root)
        parent2 = Node("parent2", None, tracker.root)

        child = Node("child", None, parent1)
        parent1.addchild(child)
        assert list(parent1.children) == [child]
        assert child.parent is parent1

        child = child.detach()
        assert child.parent is None
        assert list(parent1.children) == []

        child.parent = parent2
        parent2.addchild(child)
        assert list(parent2.children) == [child]
        assert child.parent is parent2

    def test_detach_noop(self):
        tracker = self.tracker()

        node = Node("node", None, None)
        node = node.detach()
        assert not node.parent
        assert not node.root

        parent2 = Node("parent2", None, tracker.root)
        parent2.addchild(node)

        assert list(parent2.children) == [node]
        assert node.root is tracker.root
        assert node.parent is parent2

    def test_copy_new_tracker(self):
        tracker1 = self.tracker()
        tracker2 = self.tracker()

        node1 = Node("node", "text", tracker1.root)
        tracker1.root.addchild(node1)
        node1_child = Node("nodechild", None, node1)
        node1.addchild(node1_child)

        node2 = node1.copy(tracker2.root)
        tracker2.root.addchild(node2)

        assert node2.root is tracker2.root
        assert node2.parent is tracker2.root
        node2_child = node2.children.next_neighbor
        assert node2_child.root is tracker2.root
        assert node2_child.parent is node2

        assert node1.node_type == node2.node_type
        assert node1.text == node2.text
        assert node1_child.node_type == node2_child.node_type
        assert node1_child.text == node2_child.text

    def test_copy_sibling(self):
        tracker = self.tracker()

        node = Node("node", "text", tracker.root)
        tracker.root.addchild(node)
        newnode = node.copy()
        tracker.root.addchild(newnode)

        assert list(tracker.root.children) == [node, newnode]
        assert newnode.node_type == node.node_type
        assert newnode.text == node.text
        assert newnode.parent is node.parent
        assert newnode.root is node.root

    def test_copy_newplace(self):
        tracker = self.tracker()

        parent1 = Node("parent1", None, tracker.root)
        parent2 = Node("parent2", None, tracker.root)

        node = Node("node", None, parent1)

        newnode = node.copy(parent2)
        parent2.addchild(newnode)

        assert newnode.node_type == node.node_type
        assert newnode.text == node.text
        assert newnode.parent is parent2
        assert newnode.root is tracker.root

    def test_setoption(self):
        tracker = self.tracker()

        class OptionedNode(Node):
            options = (
                Option("option1"),
                Option("option2"),
                Option("option3", int, str)
            )

            def __init__(self, *a, **k):
                super(OptionedNode, self).__init__(*a, **k)
                self.option1 = None
                self.option2 = None
                self.option3 = None

        node = OptionedNode("node", None, tracker.root)

        with pytest.raises(exceptions.LoadError):
            node.setoption("nonexistant", "error")

        node.setoption("option1", "text")
        node.setoption("option2", "othertext")
        node.setoption("option3", "2")

        assert node.option1 == "text"
        assert node.option2 == "othertext"
        assert node.option3 == 2

        with pytest.raises(ValueError):
            node.setoption("option3", "ten")

    def test_option_values(self):
        class FakeOption(object):
            def __init__(self, name, provideout):
                self.name = name
                self.incoming = None
                self.outgoing = None
                self.provideout = provideout

            def set(self, parent, value):
                self.incoming = (parent, value)

            def get(self, parent):
                self.outgoing = parent
                return self.provideout

        option1_sentinel = object()
        option1 = FakeOption("option1", (True, option1_sentinel))
        option2_sentinel = object()
        option2 = FakeOption("option2", (False, option2_sentinel))

        class OptionNode(Node):
            options = (
                option1,
                option2
            )

        tracker = self.tracker()
        node = OptionNode("node", None, tracker.root)

        node.setoption("option1", "option1_value")
        node.setoption("option2", "option2_value")

        assert node.option_values() == [
            ("option1", option1_sentinel, True),
            ("option2", option2_sentinel, False)
        ]

        assert option1.incoming == (node, "option1_value")
        assert option2.incoming == (node, "option2_value")
        assert option1.outgoing == node
        assert option2.outgoing == node

    def test_add_unexpected_child(self):
        tracker = self.tracker()
        parent = Node("parent", None, tracker.root)
        parent2 = Node("parent2", None, tracker.root)
        child = Node("child", None, parent)

        with pytest.raises(exceptions.LoadError):
            parent2.addchild(child)

    def test_str(self):
        tracker = self.tracker()
        node = Node("asdf", "fdsa", tracker.root)
        s = str(node)
        assert "asdf" in s
        assert "fdsa" in s

    def test_repr(self):
        tracker = self.tracker()
        node = Node("asdf", "fdsa", tracker.root)
        s = repr(node)
        assert "asdf" in s
        assert "fdsa" in s
        assert tracker.root.node_type in s
        assert "Node" in s

    def test_start_stop(self):
        class SimpleActivateNode(Node):
            can_activate = True

        tracker = Tracker(skeleton=False,
                nodecreator=FakeNodeCreator(SimpleActivateNode))
        tracker.root.activate(tracker.root.createchild("node1"))
        tracker.root.active_node.create("-> -node2: text")
        node3 = tracker.root.active_node.create("-> -node3: text")
        navigation.done(tracker.root)

        assert tracker.root.active_node is node3

    def test_cant_activate(self):
        node = Node("herp", None, None)
        with pytest.raises(exceptions.LoadError):
            node.start()

    def test_createchild_nodeid(self):
        tracker = self.tracker()
        tracker.root.createchild("_gennode", "text", nodeid="abcde")
        assert tracker.root.ids["abcde"].node_type == "_gennode"

    def test_ui_dictify(self):
        tracker = self.tracker()

        child = tracker.root.createchild("child", "childtext", nodeid="abcde")
        child2 = child.createchild("child2", "child2text", nodeid="12345")

        assert child.ui_dictify() == {
            "children": ["12345"],
            "type": "child",
            "text": "childtext",
            "id": "abcde"
        }

    def test_ui_graph(self):
        tracker = self.tracker()

        child = tracker.root.createchild("child", "childtext", nodeid="abcde")
        child2 = child.createchild("child2", "child2text", nodeid="12345")

        assert child.ui_graph() == {
            "abcde": {
                "children": ["12345"],
                "type": "child",
                "text": "childtext",
                "id": "abcde"
            },
            "12345": {
                "type": "child2",
                "text": "child2text",
                "id": "12345"
            },
        }

        assert tracker.root.ui_graph()["00000"]["children"] == ["abcde"]


def test_get_missing_option():
    obj = object()
    option = Option("somename")
    assert option.get(obj) == (False, None)


class TestNodeList(object):
    def test_insert_errors(self):
        nodelist = _NodeListRoot()

        node1 = GenericNode('node1')
        node2 = GenericNode('node2')
        loosenode = GenericNode('loosenode')

        nodelist.insert(node1)
        oldnext = node1._next_node
        node1._next_node = loosenode

        with pytest.raises(exceptions.ListIntegrityError):
            nodelist.insert(node2)

        node1._next_node = oldnext
        node1._prev_node = loosenode

        with pytest.raises(exceptions.ListIntegrityError):
            nodelist.insert(node2, after=nodelist)

    def test_misspecified_neighbors(self):
        nodelist = _NodeListRoot()

        node1 = GenericNode('node1')
        node2 = GenericNode('node2')
        node3 = GenericNode('node3')
        toinsert = GenericNode('toinsert')

        nodelist.insert(node1)
        nodelist.insert(node2)
        nodelist.insert(node3)

        with pytest.raises(exceptions.ListIntegrityError):
            nodelist.insert(toinsert, after=node1, before=node3)

    def test_insert_remove(self):
        nodelist1 = _NodeListRoot()
        nodelist2 = _NodeListRoot()

        node_before = GenericNode('node_before')
        node1 = GenericNode('node1')
        node_after = GenericNode('node_after')

        nodelist1.insert(node_before)
        nodelist1.insert(node1)
        nodelist1.insert(node_after)

        assert nodelist1._next_node is node_before
        assert node_before._prev_node is nodelist1
        assert node_before._next_node is node1
        assert node1._prev_node is node_before
        assert node1._next_node is node_after
        assert node_after._prev_node is node1
        assert node_after._next_node is nodelist1
        assert nodelist1._prev_node is node_after

        nodelist1.remove(node1)

        assert nodelist1._next_node is node_before
        assert node_before._prev_node is nodelist1
        assert node_before._next_node is node_after
        assert node_after._prev_node is node_before
        assert node_after._next_node is nodelist1
        assert nodelist1._prev_node is node_after

        nodelist2.insert(node1)

        assert nodelist2._next_node is node1
        assert node1._prev_node is nodelist2
        assert node1._next_node is nodelist2
        assert nodelist2._prev_node is node1

    def test_neighbors(self):
        nodelist = _NodeListRoot()
        node1 = GenericNode('node1')
        node2 = GenericNode('node2')
        nodelist.insert(node1)
        nodelist.insert(node2)

        assert nodelist.prev_neighbor is node2
        assert nodelist._prev_node is node2
        assert nodelist.next_neighbor is node1
        assert nodelist._next_node is node1

    def test_empty_neighbors(self):
        nodelist = _NodeListRoot()

        assert nodelist.prev_neighbor is None
        assert nodelist.next_neighbor is None
        assert nodelist._next_node is nodelist
        assert nodelist._prev_node is nodelist

    def test_iteration(self):
        nodelist = _NodeListRoot()

        node1 = GenericNode('node1')
        node2 = GenericNode('node2')
        node3 = GenericNode('node3')
        nodelist.insert(node1)
        nodelist.insert(node2)
        nodelist.insert(node3)

        forward = [node1, node2, node3]
        reverse = [node3, node2, node1]

        for found, expected in izip_longest(nodelist, forward):
            assert found is expected

        for found, expected in izip_longest(reversed(nodelist), reverse):
            assert found is expected


class TestRootNode(object):
    @pytest.mark.xfail
    def test_create_nonactivate(self):
        tracker = Tracker(skeleton=False)
        node = GenericActivate("herp", None, tracker.root)
        tracker.root.addchild(node)
        tracker.root.activate(node)
        tracker.root.create_after("_gennode", "honk", activate=False)
        assert tracker.root.active_node is node
        assert node.next_neighbor.text == "honk"

    @pytest.mark.xfail
    def test_create_child(self):
        tracker = Tracker(skeleton=False,
                nodecreator=FakeNodeCreator(GenericActivate))
        root = tracker.root
        node = GenericActivate("herp", "derp", tracker.root)
        root.addchild(node)
        root.activate(node)
        tracker.root.create_child("node1", "1", activate=False)
        assert root.active_node is node
        root.create_child("node2", "2")
        node1, node2 = tuple(node.children)
        assert root.active_node is node2
        assert node1.node_type == "node1"
        assert node2.node_type == "node2"

    def test_next_empty(self):
        tracker = Tracker(skeleton=False)
        node = GenericActivate("herp", None, tracker.root)
        tracker.root.addchild(node)
        tracker.root.activate(node)

        navigation.done(tracker.root)
        assert tracker.root.active_node is node

    def test_skeleton_load_integration(self, setdt):
        setdt(2013, 1, 30, 12)
        tracker = Tracker()
        tracker.deserialize("str",
            "days\n"
            "    day: today\n"
            "todo bucket\n"
            "fitness log"
        )
        today = tracker.root.active_node.text
        tracker.root.active_node.started = None
        assert match(serialize_to_str(tracker.root), (
            "days#?????\n"
            "    day#?????: {0}\n"
            "        @active\n"
            "    sleep#?????: {0}\n"
            "todo bucket#?????\n"
            "fitness log#?????\n"
        ).format(today))
        assert len(tracker.root.ids) == 6
        assert tracker.root.id == "00000"
        assert "00000" in tracker.root.ids
        assert tracker.root.ids["00000"] is tracker.root
        for depth, node in tracker.root.iter_flat_children():
            assert node.id in tracker.root.ids
            assert tracker.root.ids[node.id] is node

    def test_skeleton_day_active(self, setdt):
        setdt(2013, 1, 30, 12)
        tracker = Tracker()
        tracker.deserialize("str",
            "days\n"
            "    day#bcdef: today\n"
            "        @started: September 23, 2012 11:00 AM\n"
            "        _genactive#cdefg: something\n"
            "            @active\n"
            "    sleep#defgh: today\n"
            "todo bucket#ghijk\n"
            "fitness log#hijkl"
        )
        today = tracker.root.find("days > day: today").one()
        tracker.root.active_node.started = None
        assert match(serialize_to_str(tracker.root), (
            "days#00001\n"
            "    day#bcdef: {0}\n"
            "        @started: September 23, 2012 11:00:00 AM\n"
            "        _genactive#cdefg: something\n"
            "            @active\n"
            "    sleep#defgh: {0}\n"
            "todo bucket#ghijk\n"
            "fitness log#hijkl\n"
        ).format(today.text))
        assert set(tracker.root.ids) == {
                "00001", "bcdef", "cdefg", "defgh", "ghijk", "hijkl", "00000"}



from __future__ import unicode_literals, print_function

import re
from datetime import datetime

from treeoflife import timefmt
from treeoflife.nodes.node import Node, nodecreator
from treeoflife.exceptions import LoadError


@nodecreator("fitness log")
class FitnessLog(Node):
    toplevel = True
    allowed_children = ["weight", "waist", "calories", "workout", "_", "log"]

    def load_finished(self):
        self.root.fitness_log = self

    @classmethod
    def make_skeleton(cls, root):
        root.fitness_log = root.find("fitness log").first()
        if not root.fitness_log:
            root.fitness_log = root.createchild("fitness log")


def full_match(regex, *args):
    return re.match(regex + "$", *args)


# TODO: these names are a little wonky
@nodecreator("log")
class LogNode(Node):
    children_of = ["fitness log"]
    allowed_children = []
    preferred_parent = "< :{last} > fitness log"

    options = (
        timefmt.DatetimeOption("time"),
    )

    def __init__(self, *args, **kw):
        self.time = datetime.now()
        super(LogNode, self).__init__(*args, **kw)


class FitnessLogNode(LogNode):
    value_name = None
    value_format = None
    value_regex = None
    value_type = float

    context_name = None
    context_format = None
    context_regex = "()"

    def __init__(self, *args, **kw):
        setattr(self, self.value_name, None)
        setattr(self, self.context_name, None)

        super(FitnessLogNode, self).__init__(*args, **kw)

    @property
    def text(self):
        value = getattr(self, self.value_name)
        context = getattr(self, self.context_name)
        if value is None:
            return None

        result = self.value_format % value
        if context:
            result += self.context_format % context
        return result

    @text.setter
    def text(self, newtext):
        self.time = datetime.now()

        regex = self.value_regex + (r'(?:%s)?' % self.context_regex)
        match = full_match(regex, newtext.strip(), re.IGNORECASE)
        if match:
            value = self.value_type(match.group(1))
            setattr(self, self.value_name, value)
            context = match.group(2)
            if context and context.strip():
                setattr(self, self.context_name, context.strip())
            else:
                setattr(self, self.context_name, None)
        else:
            raise LoadError("Bad %s or %s format: %r" % (self.value_name,
                self.context_name, newtext))


class Measurement(FitnessLogNode):
    context_name = "clothes"
    context_format = " wearing %s"
    context_regex = " wearing (.+)"


@nodecreator("weight")
class Weight(Measurement):
    value_name = "weight"
    value_format = "%.03flbs"
    value_regex = (r'([0-9.]+)'
            r'(?:\s*lbs)?')


@nodecreator("waist")
class Waist(Measurement):
    value_name = "waist"
    value_format = "%.03fin"
    value_regex = (r'([0-9.]+)'
            r'(?:\s*(?:in|inch|inches|"))?')


@nodecreator("cal")
@nodecreator("cals")
@nodecreator("calories")
class Calories(FitnessLogNode):
    context_name = "food"
    context_format = " from %s"
    context_regex = " from (.+)"

    value_name = "calories"
    value_format = "%d"
    value_regex = r'([0-9]+)'

    def __init__(self, node_type, *args, **kwargs):
        if node_type == "cal" or node_type == "cals":
            node_type = "calories"
        super(Calories, self).__init__(node_type, *args, **kwargs)


@nodecreator("workout")
class Workout(FitnessLogNode):
    context_name = "activity"
    context_format = " %s"
    context_regex = " (.+)"

    value_name = "minutes"
    value_format = "%dmin"
    value_regex = (r'([0-9]+)'
            r'(?:\s*(?:min|minutes))?')


@nodecreator("tzchange")
class TZChangeNode(Node):
    text_required = True
    preferred_parent = None
    children_of = None

    options = (
        timefmt.DatetimeOption("origin_time"),
    )

    def __init__(self, *args, **kw):
        self.origin_time = datetime.now()
        Node.__init__(self, *args, **kw)



from __future__ import unicode_literals, print_function

from treeoflife import parseutil
import parsley


def test_grammar():
    class MyOtherGrammar(parseutil.Grammar):
        grammar = """
            target :arg = s('herp') -> arg + 100
        """

    class MyGrammar(parseutil.Grammar):
        grammar = """
            source :arg = othergrammar.target(arg):t ' derp' -> t + 10
        """
        bindings = {
            "othergrammar": MyOtherGrammar
        }

    assert MyGrammar("herp derp").source(1000) == 1110
    assert MyGrammar("herk derk").source(1000, optional=True) is None
    assert MyGrammar("herp derp").source(1000, optional=True) == 1110

    MyGrammar.printSource()



from __future__ import unicode_literals, print_function

from treeoflife.nodes.node import _NodeCreatorTracker
from collections import defaultdict
from fnmatch import fnmatchcase
from difflib import ndiff


def match(string, pattern):
    matches = fnmatchcase(string, pattern)
    if not matches:
        diff = ndiff(string.splitlines(1), pattern.splitlines(1))
        print("".join(diff))
    return matches


class FakeNodeCreator(_NodeCreatorTracker):
    def __init__(self, create=None):
        _NodeCreatorTracker.__init__(self)
        if create is None:
            from treeoflife.nodes.misc import GenericNode
            create = GenericNode
        self.creators = defaultdict(lambda: create)

    def exists(self, node_type):
        return True  # pragma: no cover

    def values(self):
        return [self.create]  # pragma: no cover


class _Any(object):
    def __eq__(self, other):
        return True

any = _Any()



"""
Nodes for interactive/full integration testing purposes
"""
from __future__ import unicode_literals, print_function

import datetime

from treeoflife.nodes.node import Node, nodecreator, BooleanOption
from treeoflife.nodes.tasks import BaseTask
from treeoflife import alarms


@nodecreator("_test_alarm")
class AlarmTestNode(BaseTask, alarms.NodeMixin):
    def __init__(self, *args, **kwargs):
        self.thealarm = self.alarm(self.derp)
        BaseTask.__init__(self, *args, **kwargs)

    def start(self):
        if newvalue:
            self.thealarm.delta = datetime.timedelta(seconds=10)

    def finish(self):
        pass

    def derp(self):
        listeners = self.root.tracker.listeners
        for listener in listeners:
            listener.sendmessage({
                "prompt": ["test alarm fired", self.text]
            })



from __future__ import unicode_literals, print_function

import pep8
import os
import sys

parent_dir = os.path.dirname(__file__)
root_package = os.path.dirname(parent_dir)
root = os.path.dirname(root_package)
files = [os.path.join(root, filename) for filename in [
    "activities.py",
]]
paths = [
    os.path.relpath(filename) for filename in
    [root_package] + files
]
ignore = [
    "E128",  # E128 is "under-indented"
    "E127",  # E127 is "visual indent mismatch"
    "E126",  # E126 is "over-indented"
    "E124",  # Closing bracket does not match visual indentation
    "E202",  # Whitespace before ) - used for Good Reason in test_time
    "E241",  # Too much whitespace after : or ,
    "E702",  # Multiple statements on one line (get over yourself, I know when
             #     to use this)
    "E121",  # stfu about continued comment on previous line in this file
    "E201",  # whitespace after ( in function call - used for vertical time
             # alignment
]

try:
    import __pypy__
except ImportError:
    __pypy__ = None


def test_pep8():
    styleguide = pep8.StyleGuide(paths=paths, ignore=ignore,
            show_source=True, show_pep8=False)
    if __pypy__ is None:  # pragma: no branch
        report = styleguide.check_files()
        assert not report.total_errors
    else:
        print("WARNING: did not run test due to pypy")  # pragma: no cover



from __future__ import unicode_literals, print_function

import datetime
import logging

from treeoflife.nodes.node import Node, nodecreator, BooleanOption
from treeoflife import timefmt
from treeoflife.nodes.tasks import BaseTask
from treeoflife.nodes.misc import Archived
from treeoflife import alarms
from treeoflife import searching
from treeoflife import parseutil
from treeoflife.exceptions import LoadError
#from treeoflife import alarmclock

logger = logging.getLogger(__name__)

task_types = ("day", "sleep")


class DateTask(BaseTask):
    children_of = ("days",)
    text_required = True

    @property
    def can_activate(self):
        return super(DateTask, self).can_activate and self.acceptable()

    @property
    def text(self):
        date_str = timefmt.date_to_str(self.date)
        delta = timefmt.approx_delta(datetime.datetime.now().date(), self.date)
        date_str += ' (%s, %s)' % (self.date.strftime('%A'), delta)
        return date_str

    @text.setter
    def text(self, new):
        if "(" in new:
            new = new[:new.index('(')]
            new = new.strip()

        self.date = timefmt.str_to_date(new)

    def start(self):
        BaseTask.start(self)
        if self.root.loading_in_progress:
            self.load_finished = self._post_started
        else:
            self._post_started()

    def _post_started(self):
        if self.load_finished == self._post_started:
            del vars(self)["load_finished"]
        self.post_started()

    def post_started(self):
        pass

    def ui_dictify(self, result=None):
        if result is None:
            result = {}

        prev = self.prev_neighbor
        if getattr(prev, "date", None) is None:
            result["prefix_delta"] = 0
        else:
            delta = self.date - prev.date
            result["prefix_delta"] = delta.days * 4

        return super(DateTask, self).ui_dictify(result)

    def search_texts(self):
        types, texts = BaseTask.search_texts(self)
        texts.add(timefmt.approx_delta(datetime.datetime.now().date(),
            self.date))
        texts.add(self.date.strftime('%A'))
        texts.add(timefmt.date_to_str(self.date))

        return types, texts


@nodecreator("day")
class Day(DateTask):
    def __gt__(self, other):
        return self.date > other.date

    def __lt__(self, other):
        if other.node_type == "sleep":
            return self.date <= other.date
        return self.date < other.date

    def acceptable(self):
        """
        Returns level of acceptable-ness of activating this day
        """
        now = datetime.datetime.now()
        origin = datetime.datetime.combine(self.date, datetime.time.min)
        start = origin + datetime.timedelta(hours=6)
        end = origin + datetime.timedelta(days=1)
        morning = end + datetime.timedelta(hours=6)
        if now < origin or now > morning:
            return 0
        if now >= end or now <= start:
            return 1
        return 3

    def post_started(self):
        self.parent.prep_sleep(sleep_day=self.date)


def pick_best(func, options, target):
    results = []
    for option in options:
        result = func(option)
        result = result - target
        results.append((abs(result), option))

    results = sorted(results, key=lambda x: x[0])
    result, option = results[0]

    return option


@nodecreator("sleep")
class Sleep(DateTask, alarms.NodeMixin):
    options = (
        timefmt.TimeOption("until_time"),
        timefmt.DatetimeOption("until"),
        timefmt.TimedeltaOption("amount"),
        BooleanOption("sleep_music_played")
    )

    def __init__(self, *a, **kw):
        self.wake_alarm = None  # self.alarm(self.wakeup_music)
        self.sleep_music_alarm = None  # self.alarm(self.sleep_music)
        self.canceller = None
        self.amount = None
        self.until_time = None
        self.until = None
        self.sleep_music_played = False

        DateTask.__init__(self, *a, **kw)

#    def load_finished(self):
#        assert self.config

    def __gt__(self, other):
        if other.node_type == "day":
            return self.date >= other.date
        return self.date > other.date

    def __lt__(self, other):
        return self.date < other.date

    def acceptable(self):
        # because sleep is shifted forwards
        if ((datetime.datetime.now() - datetime.timedelta(hours=12)).date()
                == self.date):
            return 2
        return 0

    def initialize(self, date, amount=None, until=None):
        self.date = date
        self.update(amount=amount, until=until)

    def _combine_until(self, until):
        assert False, "sleep node is deactivated, please don't run me"
        now = datetime.datetime.now()

        def evaluate(date):
            dt = datetime.datetime.combine(date, until)
            td = dt - now
            if td < datetime.timedelta():
                return datetime.timedelta(days=1000)
            return td

        best = pick_best(evaluate,
                set([self.date,
                    self.date + datetime.timedelta(days=1),
                    self.date + datetime.timedelta(days=2)]),
                datetime.timedelta(hours=8))
        until = datetime.datetime.combine(best, until)
        assert until > now

        return until

    def update(self, amount=None, until=None):
        return
        assert not self.active, "can't update an active sleep node"
        # can't do both at once
        assert amount is None or until is None, (
                "can't set conflicting times for sleep node")
        # if doing neither, don't erase previous
        if amount is None and until is None:
            return

        if amount is not None:
            self.amount = amount
        else:
            self.amount = None

        if isinstance(until, time):
            self.until_time = until
            until = self._combine_until(until)
        else:
            self.until_time = None

        if until is not None:
            self.until = until
        else:
            self.until = None
        logger.debug("setting %r until %r", self, self.until)

    def _adjust_times(self):
        assert False, "sleep node is deactivated, please don't run me"
        if self.amount is not None:
            self.until = datetime.datetime.now() + self.amount

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, active):
        if 0:
            assert False, "sleep node is deactivated, please don't run me"
            self._adjust_times()
            if not self.until:
                self.update(until=time(7, 0))

            self.wake_alarm.date = self.until
            self.sleep_music()

            logger.debug("marking %r active until %r", self, self.until)
        #else:
        #    self.wake_alarm.date = None
        #    logger.debug("marking %r inactive", self)
        self._active = active

    def finish(self):
        DateTask.finish(self)
        if self.canceller:
            assert False, "sleep node is deactivated, please don't run me"
            self.canceller.cancel()
        self.sleep_music_played = False

    @property
    def config(self):
        assert False, "sleep node is deactivated, please don't run me"
        config = self.root.tracker.config.setdefault("sleep", {})
        config.setdefault("evening_music", [])
        config.setdefault("wakeup_music", [])
        return config

    def sleep_music(self):
        assert False, "sleep node is deactivated, please don't run me"
        if self.sleep_music_played:
            return
        self.sleep_music_played = True

        if self.canceller is not None:
            self.canceller.cancel()

        alarmclock.set_volume(1)
        canceller = alarmclock._StopPlaying()
        canceller.add_monitor(self)
        canceller, deferred = alarmclock.play_list(
                self.config["evening_music"],
                canceller=canceller)
        self.canceller = canceller

    def wakeup_music(self):
        assert False, "sleep node is deactivated, please don't run me"
        if self.canceller is not None:
            self.canceller.cancel()

        self.fader = alarmclock.VolumeFader(alarmclock.default_curve,
                seconds=600)
        canceller, deferred = play_list(self.config["wakeup_music"],
                canceller=self.fader.canceller)
        canceller.add_monitor(self)
        self.canceller = canceller
        deferred.addCallback(self.wakeup_complete)

    def _stop_playing_trigger(self):
        assert False, "sleep node is deactivated, please don't run me"
        return self.root.tracker.root is not self.root

    def wakeup_complete(self, cancelled):
        assert False, "sleep node is deactivated, please don't run me"
        self.root.activate_next()


@nodecreator("days")
class Days(Node):
    textless = True
    toplevel = True
    allowed_children = [
        "day",
        "archived",
        "unarchive",
        "sleep"
    ]

    def __init__(self, *args, **kwargs):
        kwargs["nodeid"] = "00001"
        super(Days, self).__init__(*args, **kwargs)
        self.day_children = {}
        self.archive_date = (
                datetime.datetime.now() - datetime.timedelta(days=31)).date()

    def active_child(self):
        for parent in self.root.active_node.iter_parents():
            if parent.parent is self:
                return parent
        return None

    def active_day(self):
        child = self.active_child()
        if child is not None and child.node_type != "day":
            return None
        return child

    def active_sleep(self):
        child = self.active_child()
        if child is not None and child.node_type != "sleep":
            return None
        return child

    def prep_sleep(self, amount=None, until=None, sleep_day=None):
        current = self.active_child()
        if current.node_type == "sleep":
            raise Exception("already sleeping")

        assert current.node_type == "day"

        if sleep_day is None:
            sleep_day = (datetime.datetime.now() - datetime.timedelta(hours=4)
                    ).date()
        warn_skip = []
        sleep_node = None

        iterator = current.iter_forward()
        for node in iterator:
            if node.node_type in task_types:
                if node.date > sleep_day:
                    sleep_node = None
                    break
                elif node.node_type == "sleep" and node.date == sleep_day:
                    sleep_node = node
                    break
                else:
                    warn_skip.append(node)

        if sleep_node is None:
            sleep_node = Sleep("sleep", "today", self)
            sleep_node.initialize(date=sleep_day, amount=amount, until=until)
            self.addchild(sleep_node)
        else:
            sleep_node.update(amount=amount, until=until)

        for node in warn_skip:
            message = "WARNING: skipped %r" % node
            if not sleep_node.find("comment: %s" % message).first():
                sleep_node.createchild("comment", message)

        return sleep_node

    def sleep(self, amount=None, until=None):
        if not sleep_node.can_activate:
            raise Exception("Couldn't activate sleep node")
        prev_active = self.root.active_node
        self.root.activate(sleep_node)

        for node in prev_active.iter_parents():
            node.finish()
            if node is current:
                break

    def wakeup(self):
        sleep = self.active_sleep
        if not sleep:
            raise Exception("Already awake")

    @classmethod
    def make_skeleton(cls, root):
        root.days = root.find("days", True).first() or root.createchild('days')

        do_activate = False
        if root.active_node is None:
            do_activate = True
        else:
            for parent in root.active_node.iter_parents():
                if (parent.node_type in task_types and
                        parent.acceptable()):
                    break
            else:
                do_activate = True

        if do_activate:
            acceptables = []
            for child in root.days.children:
                if child.node_type in task_types:
                    acceptability = child.acceptable()
                    if acceptability:
                        if not child.can_activate:
                            logger.warn("node was acceptable but"
                                " could not activate: %r", child)
                            continue
                        acceptables.append((acceptability, child))

            if acceptables:
                acceptables = sorted(acceptables, key=lambda x: x[0])
                root.activate(acceptables[-1][1])
            else:
                today = root.days.createchild("day", "today")
                root.activate(today)

    def addchild(self, child, before=None, after=None):
        if child.node_type in task_types:
            before = None
            after = None
            for existing_child in reversed(self.children):
                if existing_child.node_type not in task_types:
                    continue

                if (existing_child.node_type == child.node_type and
                        existing_child.date == child.date):
                    raise LoadError("attempted to add duplicate: %r and %r" %
                            (existing_child, child))
                elif existing_child < child:
                    break
                else:
                    before = existing_child
        if child.node_type not in self.allowed_children:
            raise Exception("node %s cannot be child of %r" % (
                child._do_repr(parent=False), self))

        ret = super(Days, self).addchild(child, before=before, after=after)
        if child.node_type == "day":
            self.day_children[child.date] = child
        return ret

    def load_finished(self):
        for child in self.children:
            if child.node_type == "day" and child.date < self.archive_date:
                prev_node = child.prev_neighbor
                next_node = child.next_neighbor
                child = child.detach()
                self.addchild(Archived.fromnode(child, parent=self),
                        before=next_node, after=prev_node)

    def ui_dictify(self, result=None):
        if result is None:
            result = {}

        today = self.active_child()
        past_days = []
        future_days = []
        current = past_days
        for node in self.children:
            if node is today:
                current = future_days
            current.append(node)

        if "hidden_children" not in result and past_days:
            hidden = [child.id for child in past_days]
            result["hidden_children"] = hidden
        if "children" not in result and future_days:
            children = [child.id for child in future_days]
            result["children"] = children

        return super(Days, self).ui_dictify(result)


class _DayMatcher(searching.Matcher):
    def __init__(self, text, date=None):
        if date is None:
            self.date = timefmt.str_to_date(text)
        else:
            self.date = date
        self.text = text
        self.type = "day"
        self.create_relationship = "default"
        self.is_rigid = True

    def __call__(self, nodes, counter=None):
        for node in nodes:
            searching.tick(counter)
            if node.node_type != "day":
                continue
            if node.date != self.date:
                continue
            yield node

    def __repr__(self):
        return "_DayMatcher(%r, %r, rel=%r)" % (self.type, self.text,
                self.create_relationship)

    def copy(self):
        return _DayMatcher(self.text, self.date)


@searching.parsecreatefilters.add
def _parsehook_dayparse(queries):
    result = []
    for query in queries:
        firstseg = query.segments[0]
        if not firstseg.matcher or firstseg.separator != "children":
            continue
        matcher = firstseg.matcher
        if matcher.type is not None or matcher.text is None:
            continue
        text = matcher.text
        try:
            matcher = _DayMatcher(text)
        except (parseutil.ParseError, parseutil.EOFError):
            continue

        day_seg = searching.Segment(
                separator="children",
                pattern=('default', 'day', text),
                matcher=matcher,
                tags=set(),
                plurality=None
        )
        query = searching.Query(
                day_seg,
                *[seg.copy() for seg in query.segments[1:]]
        )
        result.append(query)

    result.extend(queries)
    return result


@searching.parsecreatefilters.add
def _parsehook_dayabs(queries):
    for query in queries:
        firstseg = query.segments[0]
        if not firstseg.matcher or firstseg.separator != "children":
            continue
        matcher = firstseg.matcher
        if matcher.type != "day" or matcher.text is None:
            continue

        if type(matcher) != _DayMatcher:
            firstseg.pattern = ('default', 'day', matcher.text)
            firstseg.matcher = _DayMatcher(matcher.text)

        days_seg = searching.Segment(
                separator="self",
                pattern=None,
                matcher=None,
                tags=set(),
                plurality=None,
                nodeid="00001"
        )
        query.segments = (days_seg,) + tuple(query.segments)
    return queries



from __future__ import unicode_literals, print_function

from datetime import datetime
import pytest

from treeoflife.tracker import Tracker, nodecreator
from treeoflife.file_storage import serialize_to_str
from treeoflife.test.util import FakeNodeCreator, match
from treeoflife import navigation
from treeoflife.nodes import tasks


class FakeDatetime(object):
    def __init__(self, time):
        self.time = time

    def now(self):
        return self.time


def test_registration():
    assert nodecreator.creators["task"] == tasks.Task


def test_task():
    tracker = Tracker(nodecreator=FakeNodeCreator(tasks.Task),
            skeleton=False)

    tracker.deserialize("str",
        "task: \xfca task\n"
        "    @started: June 7, 2010 7:00 AM"
    )


def test_active_option():
    tracker = Tracker(nodecreator=FakeNodeCreator(tasks.Task),
            skeleton=False)

    tracker.deserialize("str",
        "task: \xfca task\n"
        "    @active"
    )

    assert tracker.root.active_node.text == "\xfca task"


def test_activate_deactivate(monkeypatch, setdt):
    setdt(2012, 10, 24)
    tracker = Tracker(nodecreator=FakeNodeCreator(tasks.Task),
            skeleton=False)

    tracker.deserialize("str",
        "task: \xfc1\n"
        "    @active\n"
        "task: \xfc2\n"
        "task: \xfc3\n"
    )

    setdt(2012, 10, 25)
    navigation.done(tracker.root)
    navigation.done(tracker.root)

    assert match(tracker.serialize("str"), (
        "task#?????: \xfc1\n"
        "    @finished: 1d after October 24, 2012 12:00:00 AM\n"
        "task#?????: \xfc2\n"
        "    @finished: 0s after October 25, 2012 12:00:00 AM\n"
        "task#?????: \xfc3\n"
        "    @started: October 25, 2012 12:00:00 AM\n"
        "    @active\n"
    ))



from __future__ import unicode_literals, print_function

import json
import sys
import os
import argparse
import uuid
import shlex
import logging
from collections import defaultdict

from twisted.internet.protocol import Factory, ProcessProtocol, Protocol
import twisted.python.log
import twisted.web.static
import twisted.web.server
from txws import WebSocketFactory

from treeoflife.userinterface import (SavingInterface, command,
    generate_listing)
from treeoflife.util import Profile, setter
import treeoflife.editor_launch

logger = logging.getLogger(__name__)


@command()
def error():
    raise Exception("test")


@command()
def restart(text, ui):
    if text:
        logger.debug("restart request: %r - %s", text, text)
        args = shlex.split(str(text))
        logger.debug("restart request args: %r", args)
    else:
        args = None
    ui.restarter.restart(args)


@command()
def stop(ui):
    ui.restarter.stop()


@command()
def save(ui):
    ui.full_save()


@command()
def quit_popup(source):
    source.sendmessage({"should_quit": True})


@command()
def update(ui):
    ui.update_all()


class RemoteInterface(SavingInterface):
    max_format_depth = 3

    def __init__(self, config, restarter, *args, **keywords):
        super(RemoteInterface, self).__init__(*args, **keywords)
        self.listeners = []
        self.run_config = config
        self.restarter = restarter
        self.edit_session = None

    def deserialize(self, *a, **kw):
        SavingInterface.deserialize(self, *a, **kw)
        for listener in self.listeners:
            listener.update()

    def hide_all_clients(self):
        for listener in self.listeners:
            listener.sendmessage({"display": False})

    def update_all(self):
        for listener in self.listeners:
            listener.update()

    @setter
    def edit_session(self, session):
        self._realedit_session = session
        for listener in self.listeners:
            listener.update_editor_running()

    def show_client(self, client):
        client.sendmessage({"display": True})

    def command(self, source, line):
        if self.edit_session:
            self.edit_session.editor.command_attempted()
            return
        super(RemoteInterface, self).command(source, line)

    def _editor_finished(self, identifier):
        if identifier != self.edit_session.editor.identifier:
            logger.error("editor id mismatches, doing nothing: %r",
                            identifier)
            return
        logger.info("finishing edit instance: %r", identifier)
        with Profile("editor finished"):
            self.edit_session.editor.done()

    def _embedded_editor_finished(self, identifier, contents):
        if identifier != self.edit_session.editor.identifier:
            logger.error("editor id mismatches, doing nothing: %r",
                            identifier)
            return
        logger.info("finishing embedded edit instance: %r", identifier)
        with Profile("embedded edit finished"):
            self.edit_session.editor.done(contents)

    def errormessage(self, source, message):
        source.error(message)
        logger.error("errormessage from %r: %r", source, message)

    def displaychain(self):
        result = super(RemoteInterface, self).displaychain()
        realresult = []
        for x in result:
            realresult.append(x)
            if x.node_type == "days":
                break
        return realresult


class LineOnlyReceiver(Protocol):
    """
    Copied-and-edited version of twisted LineOnlyReceiver.
    I DON'T WANT A LENGTH LIMIT
    """
    _buffer = b''
    delimiter = b'\n'

    def dataReceived(self, data):
        """Translates bytes into lines, and calls line_received."""
        lines = (self._buffer + data).split(self.delimiter)
        self._buffer = lines.pop(-1)
        for line in lines:
            if self.transport.disconnecting:
                # this is necessary because the transport may be told to lose
                # the connection by a line within a larger packet, and it is
                # important to disregard all the lines in that packet following
                # the one that told it to close.
                return
            self.line_received(line)

    def line_received(self, line):
        """Override this for when each line is received.
        """
        raise NotImplementedError

    def send_line(self, line):
        """Sends a line to the other end of the connection.
        """
        return self.transport.writeSequence((line, self.delimiter))


class JSONProtocol(LineOnlyReceiver):
    delimiter = b"\n"

    def __init__(self, commandline, reactor):
        self.commandline = commandline
        self._is_transient_connection = False
        self.command_history = [""]
        self.command_index = 0
        self.update_timeout = None
        self.reactor = reactor

    def connectionLost(self, reason):
        if not self._is_transient_connection:
            if (self.commandline.edit_session
                    and self.commandline.edit_session.source is self):
                self.commandline.edit_session.editor.source_died()
            try:
                self.commandline.listeners.remove(self)
            except ValueError:
                logger.exception("Error removing listener from listeners")
            logger.info("connection lost: %r", reason)

    def sendmessage(self, message):
        self.send_line(json.dumps(message))

    def update(self):
        if self.update_timeout is not None and self.update_timeout.active():
            self.update_timeout.cancel()
        self.update_timeout = self.reactor.callLater(600, self.update)
        self.update_editor_running()

        try:
            reversed_displaychain = self.commandline.displaychain()[::-1]
            root = self.commandline.root
            pool = root.ui_graph()
            active_ref = getattr(root.active_node, "_px_root", None)
            todo = getattr(root, "todo", None)
            if todo is not None:
                todo_id = todo.id
            else:
                todo_id = None
            pool["ids"] = {
                "root": root.id,
                "days": "00001",
                "active": root.active_node.active_id,
                "active_ref": active_ref.id if active_ref else None,
                "todo_bucket": todo_id
            }
            self.sendmessage({
                "promptnodes": [node.id for node in reversed_displaychain],
                "pool": pool,
            })
            self.commandline.auto_save()
        except Exception:
            logger.exception("Error updating")
            try:
                self.sendmessage({
                    "prompt": ["** see console **", " ** update error **"],
                })
            except Exception:
                logger.exception("Error sending update notification")
                self.error("exception")

    def update_editor_running(self):
        self.sendmessage({
            "editor_running": self.commandline.edit_session is not None
        })

    def error(self, new_error):
        self.sendmessage({"error": new_error})

    def line_received(self, line):
        try:
            document = json.loads(line)
        except ValueError:
            logger.exception("Error loading line: %r", line)
            self.error("exception, see console")
            return

        for key, value in document.items():
            try:
                handler = getattr(self, "message_%s" % key)
            except AttributeError:
                logger.info("unrecognized message: %s = %s", key, value)
                self.error("exception, see console")
                continue
            try:
                handler(value)
            except Exception as e:
                logger.exception("error running handler")
                self.capture_error(e, "UNEXPECTED, SEE CONSOLE")

    def message_input(self, text_input):
        self.command_history[self.command_index] = text_input

    def message_navigate(self, direction):
        shift = -1 if direction == "up" else 1

        prev = self.command_index
        self.command_index += shift

        if self.command_index < 0:
            self.command_index = 0
        elif self.command_index >= len(self.command_history):
            self.command_index = len(self.command_history) - 1

        logger.info("navigate history: %s - %s", self.command_index,
                self.command_history[self.command_index])
        if prev == self.command_index:
            return

        self.sendmessage({"input": self.command_history[self.command_index]})

    def message_command(self, command):
        self.command_history[self.command_index] = command
        self.command_index = len(self.command_history)
        self.command_history.append("")
        try:
            self.commandline.command(self, command)
        except Exception as e:
            logger.exception("Error running command")
            self.capture_error(e)
        else:
            self.commandline.update_all()

    def capture_error(self, e, message=None):
        stred = str(e)
        name = type(e).__name__
        if name != "Exception" and stred:
            formatted = "%s: %s" % (name, stred)
        elif stred:
            formatted = stred
        else:
            formatted = name
        if message is not None:
            formatted = message + " " + formatted
        self.error(formatted)

    def message_display(self, is_displayed):
        pass

    def message_file_editor_finished(self, identifier):
        self._is_transient_connection = True
        self.reactor.callLater(0, self.commandline._editor_finished,
                identifier=identifier)
        self.transport.loseConnection()

    def message_embedded_editor_finished(self, info):
        self.commandline._embedded_editor_finished(
                info["identifier"], info["data"])

    def message_ui_connected(self, ignoreme):
        self._is_transient_connection = False
        try:
            self.sendmessage({
                "max_width": self.commandline.config["max_width"]
            })
            self.sendmessage({
                "display": True
            })
            self.update()
            self.commandline.listeners.append(self)
        except Exception:
            logger.exception("Error sending info on connection")


class JSONFactory(Factory):
    def __init__(self, interface, reactor):
        self.interface = interface
        self.reactor = reactor

    def buildProtocol(self, addr):
        return JSONProtocol(self.interface, self.reactor)


argparser = argparse.ArgumentParser(description="run server")
argparser.add_argument("--dev", nargs="?", dest="dev", default="false",
        const="true", type=lambda s: s.lower() == "true")
argparser.add_argument("-d", "--dir-path", default="~/.treeoflife",
        dest="path")
argparser.add_argument("-p", "--port", default=18081, dest="port", type=int)
argparser.add_argument("-l", "--log", default="server", dest="logname")
argparser.add_argument("--log-ext", default="log", dest="log_ext")
argparser.add_argument("-m", "--main-file", default="life", dest="mainfile")
argparser.add_argument("--interface", default="127.0.0.1", dest="listen_iface")
argparser.add_argument("--ignore-tests", action="store_true",
        dest="ignore_tests")
argparser.add_argument("-e", "--editor", default="embedded", dest="editor")


class Restarter(object):
    def __init__(self):
        self.should_restart = False
        self.orig_cwd = os.path.realpath(".")
        self.to_flush = [sys.stdout, sys.stderr]
        self.args = sys.argv

    def restart(self, args=None):
        reactor.stop()
        if args is not None:
            self.args = [self.args[0]] + args
        self.should_restart = True

    def stop(self):
        reactor.stop()
        self.should_restart = False

    def call(self, target, *args):
        target(self, *args)

        if self.should_restart:
            os.chdir(self.orig_cwd)
            logger.debug("restarting - args: %r", self.args)
            sys.__stdout__.write("** restarting **\n")
            for f in self.to_flush:
                f.flush()
            for f in self.to_flush:
                os.fsync(f.fileno())
            os.execv(sys.executable, [sys.executable] + self.args)


def init_sentry():
    try:
        sentryurl = open("sentryurl", "r").read().strip()
    except IOError:
        return

    from raven import Client
    from raven.handlers.logging import SentryHandler
    from raven.conf import setup_logging

    # Manually specify a client
    client = Client(
        dsn=sentryurl,
        list_max_length=256,
        string_max_length=2 ** 16,
        auto_log_stacks=True,
        timeout=1
    )
    handler = SentryHandler(client)
    handler.setLevel(logging.WARNING)

    setup_logging(handler)


def init_log(config):
    rootlogger = logging.getLogger()

    formatter = logging.Formatter('[%(asctime)s %(levelname)8s] %(name)s: '
                                            '%(message)s')

    init_sentry()

    rootlogger.setLevel(logging.DEBUG)
    logfile = open(config.logfile, "a")
    handlers = [
        logging.StreamHandler(sys.stdout),
        logging.StreamHandler(logfile)
    ]
    for handler in handlers:
        handler.setFormatter(formatter)
        rootlogger.addHandler(handler)

    twisted_observer = twisted.python.log.PythonLoggingObserver()
    twisted_observer.start()

    logger.info("logfile: %r" % config.logfile)

    return logfile


class NoCacheFile(twisted.web.static.File):
    def render_GET(self, request):
        """
        Begin sending the contents of this L{File} (or a subset of the
        contents, based on the 'range' header) to the given request.
        """

        from twisted.web.static import getTypeAndEncoding, resource, server

        # no really that's supposed to be "restat"
        self.restat(False)

        if self.type is None:
            self.type, self.encoding = getTypeAndEncoding(self.basename(),
                                              self.contentTypes,
                                              self.contentEncodings,
                                              self.defaultType)

        request.setHeader(str("Cache-Control"),
                str("no-cache, no-store, must-revalidate"))
        request.setHeader(str("Pragma"), str("no-cache"))
        request.setHeader(str("Expires"), str("0"))

        if not self.exists():
            return self.ensure_no_cache(self.childNotFound.render(request))

        if self.isdir():
            return self.redirect(request)

        request.setHeader(str('accept-ranges'), str('bytes'))

        try:
            fileForReading = self.openForReading()
        except IOError, e:
            import errno
            if e[0] == errno.EACCES:
                return resource.ForbiddenResource().render(request)
            else:
                raise

        producer = self.makeProducer(request, fileForReading)

        if request.method == str('HEAD'):
            return str('')

        producer.start()
        # and make sure the connection doesn't get closed
        return server.NOT_DONE_YET
    render_HEAD = render_GET


def main(restarter, args):
    from twisted.internet import reactor

    projectroot = os.path.os.path.abspath(os.path.join(
        os.path.dirname(__file__), b".."
    ))
    os.chdir(projectroot)
    config = argparser.parse_args(args)

    if not config.ignore_tests:
        import pytest
        if pytest.main([]) != 0:
            return

    if config.dev:
        config.path += str("_dev")
        config.logname += str("_dev")
    config.logfile = str("%s.%s") % (config.logname, config.log_ext)

    logfile = init_log(config)
    restarter.to_flush.append(logfile)

    import treeoflife.nodes.import_all

    ui = RemoteInterface(config, restarter, config.path, config.mainfile,
            reactor=reactor)
    ui.load()
    ui.config.setdefault("max_width", 500)
    logger.info("ui config: %r", ui.config)

    factory = JSONFactory(ui, reactor=reactor)
    reactor.listenTCP(config.port, factory, interface=config.listen_iface)
    reactor.listenTCP(config.port + 2,
            WebSocketFactory(factory), interface=config.listen_iface)

    # serve ui directory
    ui_dir = os.path.join(projectroot, b"ui")
    resource = NoCacheFile(ui_dir)
    static = twisted.web.server.Site(resource)
    reactor.listenTCP(config.port + 1, static,
            interface=str(config.listen_iface))
    try:
        reactor.run()
    finally:
        ui.full_save()


def _main():
    print("(pre-logging init note) sys.argv:", sys.argv)
    restarter = Restarter()
    restarter.call(main, sys.argv[1:])

if __name__ == "__main__":
    _main()



from __future__ import print_function

import ometa
ometa.FAST = True
import parsley
from parsley import unwrapGrammar
from ometa.grammar import OMeta
from ometa.runtime import (ParseError, OMetaBase, EOFError, expected,
        InputStream)
import random
import string
import sys


class CaseInsensitiveStr(str):
    def __eq__(self, other):
        return self.lower() == getattr(other, "lower", lambda: None)()


class GrammarExtensions(OMetaBase):
    def __init__(self, input, *args, **kwargs):
        super(GrammarExtensions, self).__init__(input, *args, **kwargs)

    def rule_s(self, tok):
        """
        Match and return the given string, consuming any preceding whitespace.
        """
        m = self.input
        try:
            for c in tok:
                v, e = self.exactly(c)
            return tok, e
        except ParseError, e:
            self.input = m
            raise e.withMessage(expected("string", tok))

    def exactly(self, string):
        return super(GrammarExtensions, self).exactly(
                CaseInsensitiveStr(string))

    def _match_or_none(self, rulename, *args):
        def success():
            apply_result, err = self.apply(rulename, *args)
            self.considerError(err, None)
            return apply_result, self.currentError

        def failure():
            return None, self.input.nullError()

        result, err = self._or([success, failure])
        self.considerError(err, 'donk')

        self.input = InputStream(self.input.data, len(self.input.data))

        return result, self.currentError

utility_source = """
ws = ' '*
wss = ' '+
c_wss = ','? ' '+
number = <digit+>:ds -> int(ds)
number_2dg = <digit{1,2}>:ds -> int(ds)
"""

GrammarUtilities = OMeta.makeGrammar(utility_source, "GrammarBaseUtil")\
        .createParserClass(unwrapGrammar(GrammarExtensions), {})


class _GrammarMetaclass(type):
    def __new__(cls, name, bases, dct):
        if name == "__Grammar":  # for bootstrapping
            name = "Grammar"
            return type.__new__(cls, name, bases, dct)

        assert "grammar" in dct, ("Grammar subclasses must have "
                                "'grammar' attribute")
        assert bases == (Grammar,), ("_GrammarMetaclass must only be used"
                                "from Grammar subclasses")
        source = dct["grammar"]
        bindings = dct.get("bindings", {})
        superclass = dct.get("superclass", GrammarUtilities)

        grammar_class = OMeta.makeGrammar(source, "_" + name)\
                .createParserClass(unwrapGrammar(superclass), bindings)
        dct["_grammarClass"] = grammar_class

        return type.__new__(cls, name, bases, dct)


class __Grammar(object):  # Funny name for bootstrapping
    """
    copy of parsley._GrammarWrapper, with metaclass steroids
    docstring stuff removed because docstrings should describe only
    the nonobvious when they're for the original writer
    """
    __metaclass__ = _GrammarMetaclass

    def __init__(self, input):
        self._grammar = self._grammarClass(input)
        self._input = input

    def __getattr__(self, name):
        def invokeRule(*args, **kwargs):
            """
            Invoke a Parsley rule. Passes any positional args to the rule.
            """
            optional = kwargs.get("optional", False)
            try:
                if optional:
                    ret, err = self._grammar._match_or_none(name, *args)
                else:
                    ret, err = self._grammar.apply(name, *args)
            except ParseError, e:
                err = e
            else:
                try:
                    extra, _ = self._grammar.input.head()
                except EOFError:
                    return ret
            raise err
        return invokeRule

    @classmethod
    def wraprule(cls, name, optional=False):
        def wrapper(string, *args, **kwargs):
            _optional = kwargs.get("optional", optional)
            func = getattr(cls(string), name)
            return func(*args, optional=_optional)
        return wrapper

    @classmethod
    def printSource(cls):
        print(sys.modules[cls._grammarClass.__module__].__loader__.source)


# de-bootstrap the metaclass
Grammar = __Grammar
del __Grammar



from __future__ import unicode_literals, print_function

from datetime import datetime, time, timedelta, date

import pytest

from treeoflife.test.util import match
from treeoflife.tracker import Tracker
from treeoflife.nodes.days import (Day, Sleep, Days,
        _parsehook_dayparse, _parsehook_dayabs)
from treeoflife.nodes.misc import GenericActivate
from treeoflife.exceptions import LoadError
from treeoflife.nodes import days
from treeoflife import searching
from treeoflife import navigation


def test_ordering(setdt):
    setdt(2012, 12, 22, 12)

    day1 = Day("day", "December 22, 2012", None)
    sleep1 = Sleep("sleep", "December 22, 2012", None)
    day2 = Day("day", "December 23, 2012", None)
    sleep2 = Sleep("sleep", "December 23, 2012", None)

    assert day1 < day2
    assert day2 > day1

    assert day1 < sleep1
    assert sleep1 > day1

    assert sleep1 < day2
    assert day2 > sleep1

    assert sleep1 < sleep2
    assert sleep2 > sleep1

    assert day2 < sleep2
    assert sleep2 > day2

    assert sleep2 > day1
    assert day1 < sleep2


def test_acceptable_day(setdt):
    day = Day("day", "December 22, 2012", None)

    values = {
        (2012, 12, 21, 0,  0,  0):  0,
        (2012, 12, 21, 23, 59, 59): 0,

        (2012, 12, 22, 0,  0,  0):  1,
        (2012, 12, 22, 3,  0,  0):  1,
        (2012, 12, 22, 6,  0,  0):  1,

        (2012, 12, 22, 6,  0,  1):  3,
        (2012, 12, 22, 14, 35, 0):  3,
        (2012, 12, 22, 23, 59, 59): 3,

        (2012, 12, 23, 0,  0,  0):  1,
        (2012, 12, 23, 3,  0,  0):  1,
        (2012, 12, 23, 6,  0,  0):  1,

        (2012, 12, 23, 6,  0,  1):  0,
    }
    for dt, value in values.items():
        setdt(*dt)
        assert day.acceptable() == value, str(datetime(dt))


def test_acceptable_sleep(setdt):
    setdt(2012, 1, 1)
    sleep = Sleep("sleep", "December 22, 2012", None)

    values = {
        (2012, 12, 22, 8,  0):  0,
        (2012, 12, 22, 14, 35): 2,
        (2012, 12, 23, 3,  0):  2,
        (2012, 12, 23, 11, 0):  2,
        (2012, 12, 23, 13, 0):  0,
    }

    for dt, value in values.items():
        setdt(*dt)
        assert sleep.acceptable() == value, str(datetime(dt))


class _Sortnode(GenericActivate):
    def __init__(self, sort, acceptable):
        super(_Sortnode, self).__init__("day", str(sort), None)
        self._acceptable = acceptable
        self.date = sort
        self.sort = sort

    def __gt__(self, other):
        return self.sort > other.sort

    def __lt__(self, other):
        return self.sort < other.sort

    def acceptable(self):
        return self._acceptable


def should_never_run():
    assert False, "should not have been called"


class TestMakeSkeleton(object):
    def test_acceptability_selection(self):
        tracker = Tracker(skeleton=False)
        days_node = Days("days", None, tracker.root)
        tracker.root.addchild(days_node)

        days_node.addchild(_Sortnode(1, 0))
        days_node.addchild(_Sortnode(2, 0))
        days_node.addchild(_Sortnode(3, 1))
        days_node.addchild(_Sortnode(4, 2))
        days_node.addchild(_Sortnode(5, 2))
        target = _Sortnode(6, 2)
        days_node.addchild(target)
        days_node.addchild(_Sortnode(7, 1))
        days_node.addchild(_Sortnode(8, 0))

        Days.make_skeleton(tracker.root)

        assert tracker.root.active_node is target

        assert match(tracker.serialize("str"), (
            "days#?????\n"
            "    day#?????: 1\n"
            "    day#?????: 2\n"
            "    day#?????: 3\n"
            "    day#?????: 4\n"
            "    day#?????: 5\n"
            "    day#?????: 6\n"
            "        @active\n"
            "    day#?????: 7\n"
            "    day#?????: 8\n"
        ))

    def test_acceptable_parent(self, monkeypatch):
        tracker = Tracker(skeleton=False)
        days_node = Days("days", None, tracker.root)
        tracker.root.addchild(days_node)

        days_node.addchild(_Sortnode(1, 0))
        parent = _Sortnode(2, 1)
        days_node.addchild(parent)
        days_node.addchild(_Sortnode(3, 2))

        midchild = GenericActivate("generic", None, parent)
        parent.addchild(midchild)

        target = GenericActivate("target", None, midchild)
        midchild.addchild(target)
        tracker.root.activate(target)

        monkeypatch.setattr(tracker.root, "activate", should_never_run)

        Days.make_skeleton(tracker.root)

        assert tracker.root.active_node is target
        assert match(tracker.serialize("str"), (
            "days#?????\n"
            "    day#?????: 1\n"
            "    day#?????: 2\n"
            "        generic#?????\n"
            "            target#?????\n"
            "                @active\n"
            "    day#?????: 3\n"
        ))

    def test_unacceptable_parent(self, monkeypatch):
        tracker = Tracker(skeleton=False)
        days_node = Days("days", None, tracker.root)
        tracker.root.addchild(days_node)

        days_node.addchild(_Sortnode(1, 0))
        parent = _Sortnode(2, 0)
        days_node.addchild(parent)
        target = _Sortnode(3, 1)
        days_node.addchild(target)

        midchild = GenericActivate("generic", None, parent)
        parent.addchild(midchild)

        non_target = GenericActivate("non-target", None, midchild)
        midchild.addchild(non_target)
        tracker.root.activate(non_target)

        Days.make_skeleton(tracker.root)

        assert tracker.root.active_node is target
        assert match(tracker.serialize("str"), (
            "days#?????\n"
            "    day#?????: 1\n"
            "    day#?????: 2\n"
            "        generic#?????\n"
            "            non-target#?????\n"
            "    day#?????: 3\n"
            "        @active\n"
        ))

    def test_no_acceptable_activate(self, setdt, monkeypatch):
        monkeypatch.setattr(Day, "_post_started", lambda self: None)

        tracker = Tracker(skeleton=False)
        days_node = Days("days", None, tracker.root)
        tracker.root.addchild(days_node)

        setdt(2012, 12, 22, 12)

        days_node.createchild("day", "December 19, 2012 (This should be "
                "ignored)")
        days_node.createchild("day", "December 20, 2012")
        days_node.createchild("day", "December 21, 2012")

        Days.make_skeleton(tracker.root)

        assert match(tracker.serialize("str"), (
            "days#?????\n"
            "    day#?????: December 19, 2012 (Wednesday, 3 days ago)\n"
            "    day#?????: December 20, 2012 (Thursday, 2 days ago)\n"
            "    day#?????: December 21, 2012 (Friday, yesterday)\n"
            "    day#?????: December 22, 2012 (Saturday, today)\n"
            "        @started: December 22, 2012 12:00:00 PM\n"
            "        @active\n"
        ))


def test_out_of_order(setdt):
    tracker = Tracker(skeleton=False)
    days_node = Days("days", None, tracker.root)
    tracker.root.addchild(days_node)

    setdt(2013, 12, 29, 12)

    days_node.createchild("day", "December 19, 2012")
    days_node.addchild(GenericActivate("archived", "19 a"))
    days_node.addchild(GenericActivate("archived", "19 b"))

    ignoreme = days_node.createchild("day", "December 21, 2012")
    days_node.addchild(GenericActivate("archived", "21 a"))
    days_node.addchild(GenericActivate("archived", "21 b"))

    days_node.createchild("day", "December 23, 2012", before=ignoreme)
    days_node.addchild(GenericActivate("archived", "23 a"))
    days_node.addchild(GenericActivate("archived", "23 b"))

    days_node.createchild("day", "December 22, 2012")
    days_node.addchild(GenericActivate("archived", "22 out of order a"))
    days_node.addchild(GenericActivate("archived", "22 out of order b"))

    days_node.createchild("day", "December 20, 2012")
    days_node.addchild(GenericActivate("archived", "20 out of order a"))
    days_node.addchild(GenericActivate("archived", "20 out of order b"))

    result = tracker.serialize("str")
    assert match(result, (
        "days#?????\n"
        "    day#?????: December 19, 2012 (Wednesday, *)\n"
        "    archived#?????: 19 a\n"
        "    archived#?????: 19 b\n"
        "    day#?????: December 20, 2012 (Thursday, *)\n"
        "    day#?????: December 21, 2012 (Friday, *)\n"
        "    archived#?????: 21 a\n"
        "    archived#?????: 21 b\n"

        # December 22 was moved back to before 23, leaving
        # its archived nodes behind
        "    day#?????: December 22, 2012 (Saturday, *)\n"
        "    day#?????: December 23, 2012 (Sunday, *)\n"
        "    archived#?????: 23 a\n"
        "    archived#?????: 23 b\n"

        "    archived#?????: 22 out of order a\n"
        "    archived#?????: 22 out of order b\n"
        "    archived#?????: 20 out of order a\n"
        "    archived#?????: 20 out of order b\n"
    ))


def test_bad_child():
    tracker = Tracker(skeleton=False)
    days_node = Days("days", None, tracker.root)
    tracker.root.addchild(days_node)

    with pytest.raises(Exception):
        days_node.addchild(GenericActivate("herp derp", "herk derk"))


def test_archiving(setdt):
    tracker = Tracker(skeleton=False)
    tracker.root.loading_in_progress = True
    days_node = Days("days", None, tracker.root)
    tracker.root.addchild(days_node)

    setdt(2013, 12, 22, 12)

    days_node.createchild("day", "July 19, 2012")
    days_node.createchild("day", "July 21, 2012")
    days_node.createchild("day", "July 23, 2012")
    days_node.createchild("day", "July 22, 2012")
    days_node.createchild("day", "July 20, 2012")

    days_node.load_finished()
    tracker.root.loading_in_progress = False

    assert all(node.node_type == "archived" for node in days_node.children)
    assert match("\n".join(node.text for node in days_node.children), (
        "day#?????: July 19, 2012 (Thursday, *)\n"
        "day#?????: July 20, 2012 (Friday, *)\n"
        "day#?????: July 21, 2012 (Saturday, *)\n"
        "day#?????: July 22, 2012 (Sunday, *)\n"
        "day#?????: July 23, 2012 (Monday, *)"
    ))


def test_ui_dictify(setdt, monkeypatch):
    monkeypatch.setattr(Day, "_post_started", lambda self: None)
    tracker = Tracker(skeleton=False)
    days_node = Days("days", None, tracker.root)
    tracker.root.addchild(days_node)

    setdt(2012, 12, 22, 12)

    before = [
        days_node.createchild("day", "December 19, 2012"),
        days_node.createchild("day", "December 20, 2012"),
        days_node.createchild("day", "December 21, 2012"),
    ]
    after = [
        days_node.createchild("day", "December 22, 2012"),
        days_node.createchild("day", "December 23, 2012"),
        days_node.createchild("day", "December 24, 2012"),
        days_node.createchild("day", "December 25, 2012"),
        days_node.createchild("day", "December 26, 2012"),
    ]

    Days.make_skeleton(tracker.root)

    assert days_node.ui_dictify() == {
        "children": [node.id for node in after],
        "hidden_children": [node.id for node in before],
        "text": None,
        "type": "days",
        "id": "00001",
    }


def test_ui_dictify_existing():
    tracker = Tracker(skeleton=False)
    days_node = Days("days", None, tracker.root)
    tracker.root.addchild(days_node)
    Days.make_skeleton(tracker.root)

    existing_children = object()
    existing_hidden_children = object()

    existing = {
        "children": existing_children,
        "hidden_children": existing_hidden_children,
    }

    assert days_node.ui_dictify(existing) == {
        "children": existing_children,
        "hidden_children": existing_hidden_children,
        "text": None,
        "type": "days",
        "id": "00001",
    }


def test_ui_dictify_rollover(setdt, monkeypatch):
    monkeypatch.setattr(Day, "_post_started", lambda self: None)
    tracker = Tracker(skeleton=False)
    days_node = Days("days", None, tracker.root)
    tracker.root.addchild(days_node)

    setdt(2012, 12, 22, 12)

    before = [
        days_node.createchild("day", "December 19, 2012"),
        days_node.createchild("day", "December 20, 2012"),
        days_node.createchild("day", "December 21, 2012"),
    ]
    after = [
        days_node.createchild("day", "December 22, 2012"),
        days_node.createchild("day", "December 23, 2012"),
        days_node.createchild("day", "December 24, 2012"),
        days_node.createchild("day", "December 25, 2012"),
        days_node.createchild("day", "December 26, 2012"),
    ]

    Days.make_skeleton(tracker.root)
    setdt(2012, 12, 23, 12)

    assert days_node.ui_dictify() == {
        "children": [node.id for node in after],
        "hidden_children": [node.id for node in before],
        "text": None,
        "type": "days",
        "id": "00001",
    }


def test_ui_dictify_sleepnode(setdt, monkeypatch):
    monkeypatch.setattr(Day, "_post_started", lambda self: None)
    tracker = Tracker(skeleton=False)
    days_node = Days("days", None, tracker.root)
    tracker.root.addchild(days_node)

    setdt(2012, 12, 21, 23)

    before = [
        days_node.createchild("day", "December 19, 2012"),
        days_node.createchild("day", "December 20, 2012"),
        days_node.createchild("day", "December 21, 2012"),
    ]
    after = [
        days_node.createchild("sleep", "December 21, 2012"),
        days_node.createchild("day", "December 22, 2012"),
        days_node.createchild("day", "December 23, 2012"),
        days_node.createchild("day", "December 24, 2012"),
        days_node.createchild("day", "December 25, 2012"),
        days_node.createchild("day", "December 26, 2012"),
    ]
    tracker.root.activate(after[0])

    assert days_node.ui_dictify() == {
        "children": [node.id for node in after],
        "hidden_children": [node.id for node in before],
        "text": None,
        "type": "days",
        "id": "00001",
    }


@pytest.mark.xfail
class TestSleepNode(object):
    def test_properly_initialized(self, setdt):
        setdt(2013, 1, 30, 23)
        node = Sleep("sleep", "today", None)

        assert node.wake_alarm.date is None
        assert not node.wake_alarm.called
        assert node.canceller is None
        assert node.until_time is None
        assert node.until is None
        assert node.amount is None

    @pytest.mark.parametrize(("until", "until_dt"), [
        (time(22, 50), datetime(2013, 1, 31, 22, 50)),
        (time(23,  1), datetime(2013, 1, 30, 23, 01)),
        (time( 1,  0), datetime(2013, 1, 31,  1,  0)),
        (time( 7,  0), datetime(2013, 1, 31,  7,  0)),
        (time(12,  0), datetime(2013, 1, 31,  12, 0)),
        (time(12,  0), datetime(2013, 1, 31,  12, 0)),
    ])
    def test_combine_until(self, setdt, until, until_dt):
        setdt(2013, 1, 30, 23)
        node = Sleep("sleep", "today", None)

        assert node._combine_until(until) == until_dt

    @pytest.mark.parametrize('hour', [1, 13])
    def test_day_creation(self, setdt, hour):
        setdt(2013, 1, 30, hour)
        tracker = Tracker(skeleton=False)

        tracker.deserialize("str",
            "days\n"
            "    day: today\n"
            "        @active\n"
        )

        days_node = tracker.root.children.next_neighbor
        node = days_node.find('sleep: today').one()
        assert node
        assert node.prev_neighbor.node_type == "day"

    def test_update(self, setdt):
        setdt(2013, 1, 30, 23)

        node = Sleep("sleep", "today", None)

        def _assert(up=None):
            if up:
                node_value.update(up)
            for name, value in node_value.items():
                if value is None:
                    assert getattr(node, name) is None
                else:
                    assert getattr(node, name) == value
        node_value = {}

        _assert({
            "amount": None,
            "until_time": None,
            "until": None
        })

        node._adjust_times()
        _assert()

        newuntil = datetime(2013, 1, 31, 12)
        node.update(until=newuntil)
        _assert({"until": newuntil})

        newuntil = time(7, 0)
        node.update(until=newuntil)
        _assert({
            "until": datetime(2013, 1, 31, 7, 0),
            "until_time": newuntil
        })

        node.update(amount=timedelta(hours=8))
        _assert({
            "until": None,
            "until_time": None,
            "amount": timedelta(hours=8)
        })

        node._adjust_times()
        _assert({
            "until": datetime(2013, 1, 31, 7, 0)
        })

    def test_activation_no_time(self):
        tracker = Tracker(skeleton=False)

        tracker.deserialize("str",
            "days\n"
            "    day: today\n"
            "        @active\n"
        )

        tracker.root.activate_next()


def test_searchhooks(setdt, monkeypatch):
    monkeypatch.setattr(searching, "parsecreatefilters", [
        _parsehook_dayparse,
        _parsehook_dayabs
    ])
    monkeypatch.setattr(searching, "parseonlyfilters", [])
    setdt(2014, 2, 19, 12)
    tracker = Tracker(skeleton=False)

    tracker.deserialize("str",
        "task: something\n"
        "days\n"
        "    day: today\n"
        "    day: tomorrow\n"
        "    day: September 20, 2014\n"
    )

    something = tracker.root.find("something").one()
    assert something.find("today").one().date == date(2014, 2, 19)
    assert something.find("tomorrow").one().date == date(2014, 2, 20)
    assert something.find("sept 20, 2014").one().date == date(2014, 9, 20)
    assert not something.find("invalid date sept 20, 1995").first()

    # subject to change
    assert not something.find("sept 21, 2014").first()

    assert (searching.parse("day: today")
            == searching.Queries(searching.parse("today").queries[0]))

    setdt(2014, 2, 20, 12)
    assert something.find("today").one().date == date(2014, 2, 20)


def test_duplicate_day(setdt):
    setdt(2014, 3, 19, 19, 40)

    tracker = Tracker(skeleton=False)
    tracker.deserialize("str",
        "days\n"
        "    day: today\n"
    )

    n = tracker.root.find("days").one()
    with pytest.raises(LoadError):
        n.createchild("day", "today")


def test_faroff_activate(setdt):
    setdt(2014, 3, 19, 20, 17)

    tracker = Tracker(skeleton=False)
    tracker.deserialize("str",
        "days\n"
        "    day: today\n"
        "        @active\n"
        "    day: March 25, 2014\n"
    )

    navigation.create("March 25, 2014 > task: something",
            tracker.root)

    with pytest.raises(searching.NoMatchesError):
        navigation.activate("March 25, 2014",
                tracker.root)

    with pytest.raises(searching.NoMatchesError):
        navigation.activate("March 25, 2014 > task: something",
                tracker.root)

    assert tracker.root.active_node is tracker.root.find("today").one()



from __future__ import unicode_literals, print_function

import re

from treeoflife.tracker import Tracker
from treeoflife.file_storage import serialize_to_str

from treeoflife.nodes.recordkeeping import (FitnessLog, Weight,
        Waist, Measurement, full_match)


class SimpleMeasurement(Measurement):
    value_name = "the_value"
    value_regex = "herp (1234) abc"
    value_format = "herp %d abc"


def test_measurement_noclothes():
    tracker = Tracker(skeleton=False)

    fitnesslog = FitnessLog("fitness log", None, tracker.root)
    fitnesslog._validate()
    tracker.root.addchild(fitnesslog)

    measurement = SimpleMeasurement("_", "herp 1234 abc", fitnesslog)
    measurement._validate()
    fitnesslog.addchild(measurement)

    assert measurement.the_value == 1234
    assert measurement.clothes is None


def test_measurement_withclothes():
    tracker = Tracker(skeleton=False)

    fitnesslog = FitnessLog("fitness log", None, tracker.root)
    fitnesslog._validate()
    tracker.root.addchild(fitnesslog)

    measurement = SimpleMeasurement("_",
            "herp 1234 abc wearing an office building", fitnesslog)
    measurement._validate()
    fitnesslog.addchild(measurement)

    assert measurement.the_value == 1234
    assert measurement.clothes == "an office building"


def test_measurement_text():
    tracker = Tracker(skeleton=False)

    fitnesslog = FitnessLog("fitness log", None, tracker.root)
    fitnesslog._validate()
    tracker.root.addchild(fitnesslog)

    measurement = SimpleMeasurement("_", "herp 1234 abc wearing a pot roast",
            fitnesslog)
    measurement._validate()
    fitnesslog.addchild(measurement)

    assert measurement.text == "herp 1234 abc wearing a pot roast"
    measurement.text = measurement.text
    assert measurement.text == "herp 1234 abc wearing a pot roast"


def test_measurement_time():
    tracker = Tracker(skeleton=False)

    fitnesslog = FitnessLog("fitness log", None, tracker.root)
    fitnesslog._validate()
    tracker.root.addchild(fitnesslog)

    measurement = SimpleMeasurement("_", "herp 1234 abc wearing a pot roast",
            fitnesslog, nodeid="abcde")
    measurement._validate()
    fitnesslog.addchild(measurement)

    result = serialize_to_str(measurement, is_root=False)
    assert result.startswith(
        "_#abcde: herp 1234 abc wearing a pot roast\n"
        "    @time: "
    )


def test_weight_re():
    assert full_match(Weight.value_regex, "10.5lbs").group(1) == "10.5"
    assert full_match(Weight.value_regex, "10.5 lbs").group(1) == "10.5"
    assert full_match(Weight.value_regex,
            "0123456789.0123456789 lbs").group(1) == "0123456789.0123456789"
    assert full_match(Weight.value_regex, "1 lbs").group(1) == "1"
    assert full_match(Weight.value_regex, "1").group(1) == "1"
    assert full_match(Weight.value_regex, "15.6  lbs").group(1) == "15.6"
    assert not full_match(Weight.value_regex, "1  ")
    assert not full_match(Weight.value_regex, "no numerals")
    assert not full_match(Weight.value_regex, "alpha then numerals 1234")
    assert not full_match(Weight.value_regex, "1234 numerals then alpha")
    assert not full_match(Weight.value_regex, "1234 lbs wearing something")


def test_waist_re():
    assert full_match(Waist.value_regex, "10.5in").group(1) == "10.5"
    assert full_match(Waist.value_regex, "10.5 inches").group(1) == "10.5"
    assert full_match(Waist.value_regex,
            "0123456789.0123456789 inch").group(1) == "0123456789.0123456789"
    assert full_match(Waist.value_regex, "1 inch").group(1) == "1"
    assert full_match(Waist.value_regex, "1").group(1) == "1"
    assert full_match(Waist.value_regex, "15.6  in").group(1) == "15.6"
    assert not full_match(Waist.value_regex, "1  ")
    assert not full_match(Waist.value_regex, "no numerals")
    assert not full_match(Waist.value_regex, "alpha then numerals 1234")
    assert not full_match(Waist.value_regex, "1234 numerals then alpha")
    assert not full_match(Waist.value_regex, "1234 lbs wearing something")



from __future__ import unicode_literals, print_function

from treeoflife.tracker import Tracker, nodecreator
from treeoflife import navigation
from treeoflife import searching
from treeoflife.nodes import references
from treeoflife.nodes import node
from treeoflife.file_storage import serialize_to_str

import pytest


@pytest.fixture
def tracker():
    tracker = Tracker(skeleton=False)
    return tracker


@pytest.fixture(params=[u"depends", u"reference"])
def reftype(request):
    return request.param


def _dump(nodes, getiterator=lambda x: x.children, depth=0, proxyinfo=True,
        ids=False):
    result = []
    for node in nodes:
        if ids:
            strnode = u"%s#%s: %s" % (node.node_type, node.id, node.text)
        else:
            strnode = u"%s: %s" % (node.node_type, node.text)
        if proxyinfo and isinstance(node, references.ProxyNode):
            strnode = u"proxy: " + strnode
        result.append(u" " * depth * 4 + strnode)
        result.extend(_dump(getiterator(node), getiterator, depth + 1,
            proxyinfo=proxyinfo, ids=ids))
    return result


def test_reference(tracker, reftype):
    tracker.deserialize("str",
        u"task#11111: \xfctarget\n"
        u"    task#22222: \xfcsomechild\n"
        u"        task#33333: \xfcsomechild\n"
        u"    comment#44444: \xfcderp\n"
        u"%s#aaaaa: <- \xfctarget\n" % reftype
    )

    assert len(tracker.root.find(reftype).one().children) == 2

    assert _dump(tracker.root.find(reftype), ids=True) == [
        u"%s#aaaaa: #11111" % reftype,
        u"    proxy: task#22222: \xfcsomechild",
        u"        proxy: task#33333: \xfcsomechild",
        u"    proxy: comment#44444: \xfcderp"
    ]


def test_nested_reference(tracker, reftype):
    tracker.deserialize("str",
        u"task#trgt1: \xfctarget\n"
        u"    task: \xfcsomechild\n"
        u"        task: \xfcsomechild\n"
        u"    comment: \xfcderp\n"
        u"task#trgt2: \xfcsome other thingy\n"
        u"    task: \xfcdo some work\n"
        u"        reference: << > \xfctarget\n"
        u"    task: \xfcdo some other work\n"
        u"%s: <- task\n" % reftype
    )

    x = _dump(tracker.root.find(reftype))
    y = [
        u"%s: #trgt2" % reftype,
        u"    proxy: task: \xfcdo some work",
        u"        proxy: reference: #trgt1",
        u"            proxy: task: \xfcsomechild",
        u"                proxy: task: \xfcsomechild",
        u"            proxy: comment: \xfcderp",
        u"    proxy: task: \xfcdo some other work"
    ]
    assert x == y


def test_nested_createchild(tracker, reftype):
    tracker.deserialize("str",
        u"task#trgt1: \xfctarget\n"
        u"    task: \xfcsomechild\n"
        u"        task: \xfcsomechild\n"
        u"    comment: \xfcderp\n"
        u"task#trgt2: \xfcsome other thingy\n"
        u"    task: \xfcdo some work\n"
        u"        reference: << > \xfctarget\n"
        u"    task: \xfcdo some other work\n"
        u"%s: <- task\n" % reftype
    )

    somechild = tracker.root.find(u"%s > task > "
                u"reference > \xfcsomechild" % reftype).one()
    node = somechild.createchild(u"task", u"\xfctest")
    node2 = tracker.root.find(u"task > task > reference > "
                u"task > \xfctest").one()
    assert node._px_target is node2
    assert node2._px_target is tracker.root.find(
                u"task > task > \xfctest").one()

    assert _dump(tracker.root.find("*")) == [
        u"task: \xfctarget",
        u"    task: \xfcsomechild",
        u"        task: \xfcsomechild",
        u"        task: \xfctest",
        u"    comment: \xfcderp",
        u"task: \xfcsome other thingy",
        u"    task: \xfcdo some work",
        u"        reference: #trgt1",
        u"            proxy: task: \xfcsomechild",
        u"                proxy: task: \xfcsomechild",
        u"                proxy: task: \xfctest",
        u"            proxy: comment: \xfcderp",
        u"    task: \xfcdo some other work",
        u"%s: #trgt2" % reftype,
        u"    proxy: task: \xfcdo some work",
        u"        proxy: reference: #trgt1",
        u"            proxy: task: \xfcsomechild",
        u"                proxy: task: \xfcsomechild",
        u"                proxy: task: \xfctest",
        u"            proxy: comment: \xfcderp",
        u"    proxy: task: \xfcdo some other work"
    ]


def test_createchild(tracker, reftype):
    tracker.deserialize("str",
        u"task#trgt1: \xfctarget\n"
        u"    task: \xfcsomechild\n"
        u"        task: \xfcsomechild\n"
        u"    comment: \xfcderp\n"
        u"%s: <- \xfctarget\n" % reftype
    )

    somechild = tracker.root.find(u"%s > \xfcsomechild" % reftype).one()
    somechild.createchild(u"task", u"\xfctest")

    assert _dump(tracker.root.find("*")) == [
        u"task: \xfctarget",
        u"    task: \xfcsomechild",
        u"        task: \xfcsomechild",
        u"        task: \xfctest",
        u"    comment: \xfcderp",
        u"%s: #trgt1" % reftype,
        u"    proxy: task: \xfcsomechild",
        u"        proxy: task: \xfcsomechild",
        u"        proxy: task: \xfctest",
        u"    proxy: comment: \xfcderp",
    ]


def test_parent(tracker, reftype):
    tracker.deserialize("str",
        u"task: \xfctarget 1\n"
        u"    task: \xfctarget 2\n"
        u"        task: \xfctarget 3\n"
        u"%s: <-" % reftype
    )

    proxy_3 = tracker.root.find(u"%s > * > \xfctarget 3" % reftype).one()
    proxy_2 = tracker.root.find(u"%s > \xfctarget 2" % reftype).one()
    reference = tracker.root.find(reftype).one()
    assert proxy_3.parent is proxy_2
    assert proxy_2.parent is reference


class TestProxynode(object):
    def test_activate(self, tracker, reftype, setdt):
        setdt(2014, 1, 1, 1, 1)
        tracker.deserialize("str",
            u"task: \xfctarget 1\n"
            u"    task: \xfctarget 2\n"
            u"        task: \xfctarget 3\n"
            u"%s: <-\n"
            u"days\n"
            u"    day: today\n"
            u"        reference: << > %s\n"
            u"            @active\n" % (reftype, reftype)
        )

        root = tracker.root

        # manual search just in case
        days = tracker.root.children.prev_neighbor
        day = days.children.next_neighbor
        ref = day.children.prev_neighbor

        assert root.active_node is ref

        setdt.increment(seconds=1)
        navigation.done(tracker.root)
        assert root.active_node is root.find(
                u"days > day > reference > task").one()
        setdt.increment(seconds=1)
        navigation.done(tracker.root)
        assert root.active_node is root.find(
                u"days > day > reference > task > task").one()
        setdt.increment(seconds=1)
        navigation.done(tracker.root)
        assert root.active_node is root.find(
                u"days > day > reference > task").one()
        setdt.increment(seconds=1)
        navigation.done(tracker.root)
        assert root.active_node is root.find(
                u"days > day > reference").one()
        setdt.increment(seconds=1)
        navigation.done(tracker.root)
        assert root.active_node is root.find(
                u"days > day").one()

        reference = tracker.root.find(u"days > day > reference").one()
        target_3 = tracker.root.find(u"** > \xfctarget 3").one()
        target_2 = tracker.root.find(u"** > \xfctarget 2").one()
        target_1 = tracker.root.find(u"\xfctarget 1").one()

        assert reference.started
        assert target_1.started
        assert target_2.started > target_1.started
        assert target_3.started > target_2.started
        assert target_3.finished > target_3.started
        assert target_2.finished > target_3.finished
        assert not target_1.finished
        assert reference.finished > target_2.finished

    def test_reposition(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#trgt1: \xfctarget 1\n"
            u"    task: \xfctarget 2\n"
            u"        task: \xfctarget 3\n"
            u"    task: \xfctarget 4\n"
            u"%s: <- \xfctarget 1\n" % reftype
        )

        target_3 = tracker.root.find(u"task > task > \xfctarget 3").one()
        target_4 = tracker.root.find(u"task > \xfctarget 4").one()

        proxy_3 = tracker.root.find(u"%s > task > \xfctarget 3"
                                    % reftype).one()
        proxy_4 = tracker.root.find(u"%s > \xfctarget 4" % reftype).one()

        proxy_3 = proxy_3.detach()
        proxy_3.parent = proxy_4

        proxy_4.addchild(proxy_3)

        assert _dump(tracker.root.find(u"*")) == [
            u"task: \xfctarget 1",
            u"    task: \xfctarget 2",
            u"    task: \xfctarget 4",
            u"        task: \xfctarget 3",
            u"%s: #trgt1" % reftype,
            u"    proxy: task: \xfctarget 2",
            u"    proxy: task: \xfctarget 4",
            u"        proxy: task: \xfctarget 3",
        ]

    def test_reposition_2(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#trgt1: \xfctarget 1\n"
            u"    task: \xfctarget 2\n"
            u"        task: \xfctarget 3\n"
            u"    task: \xfctarget 4\n"
            u"%s#trgt2: <-\n"
            u"reference: <-\n" % reftype
        )

        target_3 = tracker.root.find(u"task > task > \xfctarget 3").one()
        target_4 = tracker.root.find(u"task > \xfctarget 4").one()

        proxy_3 = tracker.root.find(u"%s -> "
                u"reference > task > \xfctarget 3" % reftype).one()

        detached = proxy_3.detach()
        detached.parent = target_4

        target_4.addchild(detached)
        assert target_4.find(u"*").one() is target_3

        assert _dump(tracker.root.find(u"*")) == [
            u"task: \xfctarget 1",
            u"    task: \xfctarget 2",
            u"    task: \xfctarget 4",
            u"        task: \xfctarget 3",
            u"%s: #trgt1" % reftype,
            u"    proxy: task: \xfctarget 2",
            u"    proxy: task: \xfctarget 4",
            u"        proxy: task: \xfctarget 3",
            u"reference: #trgt2",
            u"    proxy: task: \xfctarget 2",
            u"    proxy: task: \xfctarget 4",
            u"        proxy: task: \xfctarget 3",
        ]

    def test_reposition_3(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#trgt1: \xfctarget 1\n"
            u"    task: \xfctarget 2\n"
            u"    task: \xfctarget 3\n"
            u"%s: <- \xfctarget 1\n" % reftype
        )

        target_2 = tracker.root.find(u"task > \xfctarget 2").one()
        target_3 = tracker.root.find(u"task > \xfctarget 3").one()

        proxy_2 = tracker.root.find(u"%s > \xfctarget 2" % reftype).one()

        detached = proxy_2.detach()
        detached.parent = target_3

        target_3.addchild(detached)
        assert target_3.find(u"*").one() is target_2

        assert _dump(tracker.root.find(u"*")) == [
            u"task: \xfctarget 1",
            u"    task: \xfctarget 3",
            u"        task: \xfctarget 2",
            u"%s: #trgt1" % reftype,
            u"    proxy: task: \xfctarget 3",
            u"        proxy: task: \xfctarget 2",
        ]

    def test_copy_1(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#trgt1: \xfctarget 1\n"
            u"    task: \xfctarget 2\n"
            u"        task: \xfctarget 3\n"
            u"    task: \xfctarget 4\n"
            u"%s: <- \xfctarget 1\n" % reftype
        )

        target_3 = tracker.root.find(u"task > task > \xfctarget 3").one()
        target_4 = tracker.root.find(u"task > \xfctarget 4").one()

        proxy_3 = tracker.root.find(reftype + u" > task > \xfctarget 3").one()
        proxy_4 = tracker.root.find(reftype + u" > \xfctarget 4").one()

        target_3_new = proxy_3.copy()
        target_3_new.parent = proxy_4

        proxy_4.addchild(target_3_new)

        assert _dump(tracker.root.find(u"*")) == [
            u"task: \xfctarget 1",
            u"    task: \xfctarget 2",
            u"        task: \xfctarget 3",
            u"    task: \xfctarget 4",
            u"        task: \xfctarget 3",
            u"%s: #trgt1" % reftype,
            u"    proxy: task: \xfctarget 2",
            u"        proxy: task: \xfctarget 3",
            u"    proxy: task: \xfctarget 4",
            u"        proxy: task: \xfctarget 3",
        ]

    def test_copy_2(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#trgt1: \xfctarget 1\n"
            u"    task: \xfctarget 2\n"
            u"        task: \xfctarget 3\n"
            u"    task: \xfctarget 4\n"
            u"%s: <- \xfctarget 1\n" % reftype
        )

        target_3 = tracker.root.find(u"task > task > \xfctarget 3").one()
        target_4 = tracker.root.find(u"task > \xfctarget 4").one()

        proxy_3 = tracker.root.find(reftype + u"> task > \xfctarget 3").one()
        proxy_4 = tracker.root.find(reftype + u"> \xfctarget 4").one()

        target_3_new = proxy_3.copy(parent=proxy_4)

        target_4.addchild(target_3_new)

        assert _dump(tracker.root.find(u"*")) == [
            u"task: \xfctarget 1",
            u"    task: \xfctarget 2",
            u"        task: \xfctarget 3",
            u"    task: \xfctarget 4",
            u"        task: \xfctarget 3",
            reftype + u": #trgt1",
            u"    proxy: task: \xfctarget 2",
            u"        proxy: task: \xfctarget 3",
            u"    proxy: task: \xfctarget 4",
            u"        proxy: task: \xfctarget 3",
        ]

    def test_creation(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#trgt1: \xfctarget 1\n"
            u"    task: \xfctarget 2\n"
            u"%s: <-" % reftype
        )

        target_2 = tracker.root.find(reftype + u" > ").one()

        target_2.createchild(u"task", u"\xfctest")

        assert _dump(tracker.root.find(u"*")) == [
            u"task: \xfctarget 1",
            u"    task: \xfctarget 2",
            u"        task: \xfctest",
            reftype + u": #trgt1",
            u"    proxy: task: \xfctarget 2",
            u"        proxy: task: \xfctest"
        ]

    def test_whacky_reposition_setparent(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#toref: \xfcto reference\n"
            u"    task: \xfcdummy\n"
            u"        task: \xfcmove me\n"
            u"    task: \xfcparent\n"
            u"%s: <- \xfcto reference\n" % reftype
        )

        proxy = tracker.root.find(reftype + u" > task > \xfcmove me").one()
        parent = tracker.root.find(reftype + u"> task: \xfcparent").one()
        move_me = proxy._px_target

        move_me.detach()

        # this is to make sure assigning to parent doesn't blow things up
        proxy.parent = parent
        assert proxy.parent is parent

        parent.addchild(proxy)

        assert _dump(tracker.root.children) == [
            u"task: \xfcto reference",
            u"    task: \xfcdummy",
            u"    task: \xfcparent",
            u"        task: \xfcmove me",
            reftype + u": #toref",
            u"    proxy: task: \xfcdummy",
            u"    proxy: task: \xfcparent",
            u"        proxy: task: \xfcmove me",
        ]

    def test_whacky_reposition_noneparent(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#toref: \xfcto reference\n"
            u"    task: \xfcdummy\n"
            u"        task: \xfcmove me\n"
            u"    task: \xfcparent\n"
            u"%s: <- \xfcto reference\n" % reftype
        )

        proxy = tracker.root.find(reftype + u"> task > \xfcmove me").one()
        parent = tracker.root.find(reftype + u"> task: \xfcparent").one()
        move_me = proxy._px_target

        move_me.detach()

        # this is to make sure assigning to parent doesn't blow things up
        proxy.parent = None
        assert proxy.parent is None

        parent.addchild(proxy)

        assert _dump(tracker.root.children) == [
            u"task: \xfcto reference",
            u"    task: \xfcdummy",
            u"    task: \xfcparent",
            u"        task: \xfcmove me",
            reftype + u": #toref",
            u"    proxy: task: \xfcdummy",
            u"    proxy: task: \xfcparent",
            u"        proxy: task: \xfcmove me",
        ]

    def test_set_text(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#toref: \xfcto reference\n"
            u"    task: \xfcto rename\n"
            u"%s: <-" % reftype
        )

        to_rename = tracker.root.find(reftype + u' > \xfcto rename').one()
        to_rename.text = u"\xfcrenamed"

        assert _dump(tracker.root.children) == [
            u"task: \xfcto reference",
            u"    task: \xfcrenamed",
            reftype + ": #toref",
            u"    proxy: task: \xfcrenamed"
        ]

    def test_export(self, tracker, reftype):
        tracker.deserialize("str",
            u"task: \xfcto reference\n"
            u"    task#aaaaa: \xfcparent\n"
            u"        task: \xfcchild 1\n"
            u"        task: \xfcchild 2\n"
            u"        task: \xfcchild 3\n"
            u"%s: <-" % reftype
        )

        from treeoflife.file_storage import serialize

        result = serialize(tracker.root.find(reftype + u" > \xfcparent").one())
        assert result == [
            u"task#aaaaa: \xfcparent"
        ]

    def test_no_options(self, tracker, reftype):
        tracker.deserialize("str",
            u"task: \xfcto reference\n"
            u"    task: \xfctarget\n"
            u"%s: <-" % reftype
        )

        target = tracker.root.find(reftype + u"> \xfctarget").one()

        with pytest.raises(AttributeError) as excinfo:
            target.options

        assert "do not have" in excinfo.value.message


class TestRefnode(object):
    def test_creation(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#abcde: \xfctarget 1\n"
            u"{}: <-".format(reftype)
        )

        ref = tracker.root.find(reftype).one()

        ref.createchild(u"task", u"\xfctest")

        assert _dump(tracker.root.find(u"*")) == [
            u"task: \xfctarget 1",
            u"    task: \xfctest",
            reftype + u": #abcde",
            U"    proxy: task: \xfctest"
        ]

    def test_unfinish(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#abcde: \xfctarget 1\n"
            u"    task: \xfctarget 2\n"
            u"days\n"
            u"    day: today\n"
            u"        @active\n"
            u"        {}: << >\n".format(reftype)
        )

        navigation.done(tracker.root)
        refnode = tracker.root.find(u"** > " + reftype).one()
        assert tracker.root.active_node is refnode
        assert refnode.started
        assert not refnode.finished
        assert _dump([refnode]) == [
            reftype + u": #abcde",
            u"    proxy: task: \xfctarget 2"
        ]
        navigation.finish(u"<", tracker.root)
        target = tracker.root.find(u"task: \xfctarget 1").one()
        if reftype == u"depends":
            assert target.finished
        else:
            assert not target.finished
        assert refnode.finished
        assert _dump([refnode]) == [
            reftype + u": #abcde"
        ]
        assert tracker.root.active_node.node_type == u"day"

        navigation.forceactivate(u">", tracker.root)
        assert tracker.root.active_node is refnode
        assert refnode.started
        assert not refnode.finished
        assert not target.finished
        assert _dump([refnode]) == [
            reftype + u": #abcde",
            u"    proxy: task: \xfctarget 2"
        ]

    def test_unfinish_notinitialized(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#abcde: \xfctarget 1\n"
            u"    task: \xfctarget 2\n"
            u"days\n"
            u"    day: today\n"
            u"        @active\n"
            u"        {}: << >\n"
            u"            @started\n"
            u"            @finished\n".format(reftype)
        )

        refnode = tracker.root.find(u"** > " + reftype).one()
        assert refnode.finished
        assert _dump([refnode]) == [
            reftype + u": " + refnode.text
        ]
        assert tracker.root.active_node.node_type == u"day"
        target = tracker.root.find(u"task: \xfctarget 1").one()
        if reftype == u"depends":
            assert target.finished
        else:
            assert not target.finished

        navigation.forceactivate(u">", tracker.root)
        assert tracker.root.active_node is refnode
        assert refnode.started
        assert not refnode.finished
        assert not refnode._px_target.finished
        assert _dump([refnode]) == [
            reftype + u": #abcde",
            u"    proxy: task: \xfctarget 2"
        ]

    def test_unfinish_targetfinished(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#abcde: \xfctarget 1\n"
            u"    @started\n"
            u"    @finished\n"
            u"    task: \xfctarget 2\n"
            u"days\n"
            u"    day: today\n"
            u"        @active\n"
            u"        {}: << >\n"
            u"            @started\n"
            u"            @finished\n".format(reftype)
        )

        refnode = tracker.root.find(u"** > " + reftype).one()
        assert refnode.finished
        assert _dump([refnode]) == [
            reftype + u": " + refnode.text
        ]
        assert tracker.root.active_node.node_type == u"day"
        target = tracker.root.find(u"task: \xfctarget 1").one()
        assert target.finished

        navigation.forceactivate(u">", tracker.root)
        assert tracker.root.active_node is refnode
        assert refnode.started
        assert not refnode.finished
        assert not refnode._px_target.finished
        assert _dump([refnode]) == [
            reftype + u": #abcde",
            u"    proxy: task: \xfctarget 2"
        ]

    def test_initial_state(self, tracker, reftype):
        creator = tracker.nodecreator.creators[reftype]
        ref = creator(reftype, u"<-")

        assert ref._px_target is None
        assert ref.children.length == 0

        tracker.deserialize("str",
            u"task: \xfctarget\n"
            u"    task: \xfcchild\n"
        )

        tracker.root.addchild(ref)

        assert ref._px_target is tracker.root.find(u"task").one()

    def test_export(self, tracker, reftype):
        tree = (
            u"task#55555: \xfctarget\n"
            u"    task#44444: \xfcsomechild\n"
            u"        task#33333: \xfcsomechild\n"
            u"    comment#22222: \xfcderp\n"
            u"{}#11111: %s\n".format(reftype)
        )
        tracker.deserialize("str", tree % u"<- \xfctarget")

        assert tracker.serialize("str") == tree % u"#55555"

    def test_removechild(self, tracker, reftype):
        tracker.deserialize("str",
            u"task#abcde: \xfcreferenced\n"
            u"    task: \xfctoremove\n"
            u"{}: <-\n".format(reftype)
        )
        reference = tracker.root.find(reftype).one()

        toremove = reference.find(u"task").one()
        reference.removechild(toremove)
        assert _dump(tracker.root.children) == [
            u"task: \xfcreferenced",
            u"{}: #abcde".format(reftype)
        ]


def test_child_loop(tracker, reftype):
    tracker.deserialize("str",
        u"task: a\n"
        u"    task: b\n"
        u"        task: e\n"
        u"        task: f\n"
        u"        task: g\n"
        u"    task: c\n"
        u"    task: d\n"
        u"%s: <-" % reftype
    )

    a = tracker.root.find(reftype).one()
    b = a.find(u"b").one()
    c = a.find(u"c").one()
    d = a.find(u"d").one()
    e = a.find(u"b > e").one()
    f = a.find(u"b > f").one()
    g = a.find(u"b > g").one()

    assert a.children._next_node is b
    assert a.children is b._prev_node
    assert c is b._next_node
    assert c._prev_node is b
    assert c._next_node is d
    assert c is d._prev_node
    assert a.children is d._next_node
    assert a.children._prev_node is d

    assert b.children._next_node is e
    assert b.children is e._prev_node
    assert f is e._next_node
    assert f._prev_node is e
    assert f._next_node is g
    assert f is g._prev_node
    assert b.children is g._next_node
    assert b.children._prev_node is g


class TestSearchCreateIntegration(object):
    def test_proxied_create(self, tracker, reftype):
        tracker.deserialize("str",
            u"task: target\n"
            u"%s: <-" % reftype
        )
        creator = searching.parse_create(reftype + u" > task: test")
        node = creator(tracker.root)
        proxy = tracker.root.find(reftype + u"> test").one()
        assert node is proxy

    def test_mini_child_loop(self, tracker, reftype):
        tracker.deserialize("str",
            "task: target\n"
            "%s: <-" % reftype
        )
        creator = searching.parse_create(reftype + u"> task: test")
        node = creator(tracker.root)
        assert node

        proxy_task = tracker.root.find(reftype + u"> test").one()
        target_task = tracker.root.find(u"target > test").one()
        reference = tracker.root.find(reftype).one()
        target = tracker.root.find(u"target").one()

        unwrap = reference.unwrap
        wrap = reference.get_proxy

        assert unwrap(reference) is target
        assert wrap(target) is reference
        assert unwrap(proxy_task) is target_task
        assert wrap(target_task) is proxy_task

        assert target_task._next_node is target.children
        assert target_task._prev_node is target.children
        assert proxy_task._next_node is reference.children
        assert proxy_task._prev_node is reference.children

        assert target_task.parent is target
        assert proxy_task.parent is reference


def test_simple_interaction(tracker, reftype):
    tracker.deserialize("str",
        u"task: target\n"
        u"days\n"
        u"    day: today\n"
        u"        @started\n"
        u"        %s: << > target\n"
        u"            @active\n" % reftype
    )

    navigation.createauto(u"task: test", tracker.root)

    active_node = tracker.root.active_node
    proxy = tracker.root.find(u"days > today > " + reftype + u" > test").one()
    target = tracker.root.find(u"target > test").one()
    assert active_node is proxy
    assert proxy is not target
    assert proxy._px_target is target
    assert proxy._px_root.get_proxy(target) is proxy

    navigation.createfinish(u"-> task: test 2", tracker.root)

    target = tracker.root.find(u"target").one()
    assert target.started
    first_child = target.children.next_neighbor
    assert first_child.started
    assert first_child.finished
    second_child = first_child.next_neighbor
    assert second_child.started
    assert not second_child.finished

    proxy_active = tracker.root.find(u"days > day > "
            + reftype + u"> test 2").one()
    assert proxy_active is tracker.root.active_node


def test_another_interaction(tracker, reftype):
    tracker.deserialize("str",
        u"task: target\n"
        u"days\n"
        u"    day: today\n"
        u"        @started\n"
        u"        %s: << > target\n"
        u"            @active\n" % reftype
    )

    navigation.createauto(u"task: test 1", tracker.root)
    navigation.createfinish(u"< > +task: test 2", tracker.root)

    target = tracker.root.find(u"target").one()

    assert target.started
    first_child = target.children.next_neighbor
    assert first_child.started
    assert first_child.finished
    second_child = first_child.next_neighbor
    assert second_child.started
    assert not second_child.finished

    proxy_active = tracker.root.find(u"days > day > "
            + reftype + u" > test 2").one()
    assert proxy_active is tracker.root.active_node


def test_str(tracker, reftype):
    tracker.deserialize("str",
        u"task#abcde: target\n"
        u"    comment: to test against\n"
        u"%s: <-\n" % reftype
    )

    proxy = tracker.root.find(reftype + u" > comment").one()
    proxy_target = proxy._px_target
    assert str(proxy) == str(proxy_target)

    ref = tracker.root.find(reftype).one()
    assert str(ref) == reftype + u": target"

    if reftype == u"reference":
        ref2 = references.Reference(u"reference", u"<- task: target")
        assert str(ref2) == u"reference: <- task: target"
        assert ref2.text == u"<- task: target"

        tracker.root.addchild(ref2)
        assert str(ref2) == reftype + u": target"
        assert ref2.text == u"#abcde"
    elif reftype == u"depends":
        for x in [u"depends", u"dep", u"depend"]:
            ref2 = references.Depends(x, u"<- task: target")
            assert str(ref2) == u"depends: <- task: target"

            tracker.root.addchild(ref2)
            assert str(ref2) == reftype + u": target"
    else:  # pragma: no cover
        assert False


def test_recursion(tracker, reftype):
    tracker.deserialize("str",
        u"task#abcde: target\n"
        u"    task: something\n"
        u"        %s: < target\n" % reftype
    )

    assert _dump(tracker.root.children) == [
        u"task: target",
        u"    task: something",
        u"        %s: #abcde" % reftype,
        u"            proxy: task: something",
        u"                proxy: %s: <recursing>" % reftype,
    ]


def test_ui_dictify(tracker, reftype):
    tracker.deserialize("str",
        (u"task#targt: target\n"
        u"    task#abcde: child\n"
        u"%s#rfrnc: <-\n"
        u"%s#rfrn2: <-\n"
        ) % (reftype, reftype)
    )

    child = tracker.root.find(reftype + u" > child").one()
    assert child.ui_dictify() is None
    assert tracker.root.ui_graph() == {
        "rfrn2": {
            "type": reftype,
            "text": "#rfrnc",
            "children": ["abcde"],
            "id": "rfrn2",
            "active_id": "targt",
            "finished": False,
            "started": False,
            "target": "rfrnc"
        },
        "rfrnc": {
            "type": reftype,
            "text": "#targt",
            "children": ["abcde"],
            "id": "rfrnc",
            "active_id": "targt",
            "finished": False,
            "started": False,
            "target": "targt"
        },
        "abcde": {
            "type": "task",
            "text": "child",
            "id": "abcde",
            "active_id": "abcde",
            "finished": False,
            "started": False,
        },
        "targt": {
            "type": "task",
            "text": "target",
            "id": "targt",
            "active_id": "targt",
            "finished": False,
            "started": False,
            "children": ["abcde"]
        },
        "00000": {
            "type": "life",
            "text": None,
            "id": "00000",
            "children": ["targt", "rfrnc", "rfrn2"]
        }
    }


def test_no_query_finished(tracker, monkeypatch):
    # this one does _not_ apply to depends, hence no reftype
    # depends has to do a query to make sure the target is marked
    monkeypatch.setattr(references.Reference, "find", None)
    tracker.deserialize("str",
        u"reference: <-\n"
        u"    @started\n"
        u"    @finished"
    )


def test_reference_finished(tracker, reftype):
    tracker.deserialize("str",
        u"task: finished node\n"
        u"    @started: June 1, 2013 1:00 am\n"
        u"    @finished: June 1, 2013 2:00 am\n"
        u"%s: <-" % reftype
    )

    reference = tracker.root.find(reftype).one()
    assert not reference.started
    assert reference.finished


def test_depends_finished_node(tracker, reftype):
    tracker.deserialize("str",
        u"task: finished node\n"
        u"    @started: June 1, 2013 1:00 am\n"
        u"    @finished: June 1, 2013 2:00 am\n"
        u"%s: <-\n"
        u"    @finished: June 2, 2013 2:00 am" % reftype
    )

    reference = tracker.root.find(reftype).one()
    target = tracker.root.find(u"task").one()
    assert not reference.started
    assert reference.finished
    assert reference.finished > target.finished


def test_reference_started(tracker, reftype):
    tracker.deserialize("str",
        u"task: finished node\n"
        u"%s: <-\n"
        u"    @started: June 1, 2013 1:00 am\n" % reftype
    )

    reference = tracker.root.find(reftype).one()
    target = reference._px_target
    assert reference.started == target.started
    assert not reference._px_didstart


def test_depends_propogate_finished(tracker):
    tracker.deserialize("str",
        u"task: finished node\n"
        u"depends: <-\n"
        u"    @started: June 1, 2013 1:00 am\n"
        u"    @finished: June 2, 2013 1:00 am\n"
    )

    dep = tracker.root.find(u"depends").one()
    target = tracker.root.find(u"task").one()
    assert dep._px_target is None
    assert dep.finished == target.finished
    assert not dep._px_didfinish


def test_reference_category(tracker, reftype):
    tracker.deserialize("str",
        u"category: stuff\n"
        u"    task: stuff\n"
        u"days\n"
        u"    day: today\n"
        u"        %s: << > category" % reftype
    )

    ref = tracker.root.find(u'** > ' + reftype).one()
    tracker.root.activate(ref)

    assert ref.started


def test_reference_category_existingtime(tracker, reftype):
    tracker.deserialize("str",
        u"category: stuff\n"
        u"    task: stuff\n"
        u"days\n"
        u"    day: today\n"
        u"        %s: << > category\n"
        u"            @started: December 19, 1994 11:55 PM" % reftype
    )

    ref = tracker.root.find(u'** > ' + reftype).one()

    assert ref.started


def test_unwriteable_started(tracker, monkeypatch, reftype):
    called = []

    class NonTaskNode(node.Node):
        def __setattr__(self, name, value):
            if name == "started":
                called.append(True)
                raise AttributeError
            node.Node.__setattr__(self, name, value)

    monkeypatch.setitem(tracker.nodecreator.creators,
            "nontask", NonTaskNode)

    tracker.deserialize("str",
        u"nontask: something\n"
        u"    task: something else\n"
        u"        task: something else 3\n"
        u"    task: something else 2\n"
        u"days\n"
        u"    day: today\n"
        u"        %s: << >\n"
        u"            @started\n" % reftype
    )

    ref = tracker.root.find(u"days > day > " + reftype).one()

    assert called[0]
    assert ref.started
    assert not hasattr(ref._px_target, "started")


def test_unwriteable_finished(tracker, monkeypatch):
    called = []

    class NonTaskNode(node.Node):
        def __setattr__(self, name, value):
            if name == "finished":
                called.append(True)
                raise AttributeError
            node.Node.__setattr__(self, name, value)

    monkeypatch.setitem(tracker.nodecreator.creators,
            "nontask", NonTaskNode)

    tracker.deserialize("str",
        u"nontask: something\n"
        u"    task: something else\n"
        u"        task: something else 3\n"
        u"    task: something else 2\n"
        u"days\n"
        u"    day: today\n"
        u"        depends: << >\n"
        u"            @started\n"
        u"            @finished\n"
    )

    ref = tracker.root.find(u"days > day > depends").one()

    assert called[0]
    assert ref.finished
    assert not hasattr(ref._px_target, "finished")


# search contexts or search quoting are required to make references useful
# test finish with behavior being called
# test target nodes with no finished or started slots



from __future__ import unicode_literals, print_function

from treeoflife.exceptions import LoadError, ErrorContext


def test_loaderror():
    assert str(LoadError("derp")) == "derp"

    error_context = ErrorContext()
    error = LoadError("herk")
    error.error_context = error_context

    assert str(error) == error.message
    assert str(error) == "At unknown line: herk"

    error_context.line = 0
    assert str(error) == "At line 1: herk"

    error_context.line = "error_cause"
    assert str(error) == "herk"



from __future__ import unicode_literals, print_function

import logging
import os
import tempfile as _tempfile
import time
import functools
import __builtin__

logger = logging.getLogger(__name__)


def tempfile():
    fd, tmp = _tempfile.mkstemp()
    os.close(fd)
    return tmp


class HandlerDict(object):
    name = "handlers"
    autodetect = True

    def __init__(self):
        setattr(self, self.name, {})

    def add(self, name=None):
        assert getattr(name, "__call__", None) is None

        def _inner(func):
            _name = name
            if _name is None:
                if self.autodetect:
                    _name = func.__name__
                else:
                    raise Exception("must provide name")
            getattr(self, self.name)[_name] = func
            return func
        return _inner


class HandlerList(list):
    def add(self, func):
        self.append(func)
        return func


class Profile(object):
    def __init__(self, name):
        self.name = name

    def __enter__(self):
        self.started = time.time()

    def __exit__(self, x, y, z):
        self.finished = time.time()
        logger.debug("Profile %r finished in %r", self.name,
                self.finished - self.started)


def memoize(obj):
    cache = {}
    invalidate = [0]

    @functools.wraps(obj)
    def wrapper(*a, **kw):
        fs_kw = frozenset(kw.items())
        extra = int(time.time()) / 20
        if invalidate[0] != extra:
            cache.clear()
            invalidate[0] = extra

        try:
            return cache[a, fs_kw]
        except KeyError:
            z = cache[a, fs_kw] = obj(*a, **kw)
            return z
    wrapper.cache = cache
    return wrapper

template = """
@property
def {0}(self):
    return self._hack_{0}

@{0}.setter
def {0}(self, newvalue):
    import logging
    logger = logging.getLogger("_monitor.{0}")
    logger.debug("writing: %s", newvalue)
    import traceback
    f = "".join(traceback.format_stack())
    print(f)
    self._hack_{0} = newvalue
"""


def setter(func):
    name = func.__name__

    @functools.wraps(func)
    def get(self):
        return getattr(self, "_real" + name)

    return property(get, func)


def _monitor(name):
    import inspect
    frames = inspect.stack()
    f = frames[1][0]
    render = template.format(name)
    exec(render, f.f_globals, f.f_locals)


def hasattr_(obj, name):
    """
    Injected hasattr that only catches AttributeError
    """
    sentinel = object()
    return getattr(obj, name, sentinel) is not sentinel

hasattr_.is_good_hasattr = True

__builtin__.hasattr = hasattr_



from __future__ import unicode_literals, print_function

import traceback

from twisted.internet.defer import Deferred

from treeoflife.exceptions import LoadError, ErrorContext
from treeoflife.nodes.node import TreeRootNode, nodecreator
from treeoflife.file_storage import loaders, serializers


class Tracker(object):
    def __init__(self, skeleton=True, nodecreator=nodecreator,
            roottype=TreeRootNode):
        self.make_skeleton = skeleton
        self.nodecreator = nodecreator
        self.config = {}

        self.roottype = roottype

        self.root = self.roottype(self, self.nodecreator,
                loading_in_progress=False)
        if self.make_skeleton:
            self.root.make_skeleton()

    def deserialize(self, format, reader):
        self.root = root = self.roottype(self, self.nodecreator,
                loading_in_progress=True)
        root.loading_in_progress = True
        stack = []
        lastnode = root
        lastindent = -1
        metadata_allowed_here = False

        error_context = ErrorContext()  # mutable thingy

        try:
            parser = loaders.handlers[format](reader)
            parser.error_context = error_context
            for indent, is_metadata, nodeid, node_type, text in parser:
                if node_type == "":
                    indent = lastindent
                    if lastnode is not None and lastnode.node_type != "":
                        indent += 1
                if indent > lastindent:
                    if indent > lastindent + 1:
                        raise LoadError("indented too far")
                    stack.append(lastnode)
                    metadata_allowed_here = True
                elif indent < lastindent:
                    stack = stack[:int(indent) + 1]
                lastindent = indent

                parent = stack[-1]

                if is_metadata:
                    if not metadata_allowed_here:
                        raise LoadError('metadata in the wrong place')
                    parent.setoption(node_type, text)
                    lastnode = None
                else:
                    if node_type != "-":
                        metadata_allowed_here = False

                    node = self.nodecreator.create(node_type, text, parent,
                            nodeid=nodeid)
                    if node is not None:
                        parent.addchild(node)
                    lastnode = node
        except LoadError as e:
            e.error_context = error_context
            raise
        except Exception as e:
            new_e = LoadError("UNHANDLED ERROR:\n %s" % traceback.format_exc())
            new_e.error_context = error_context
            raise new_e

        if self.make_skeleton:
            root.make_skeleton()

        root.load_finished()
        for depth, node in root.iter_flat_children():
            node.load_finished()

        # enable instant load_finished() on node creation
        root.loading_in_progress = False

    def serialize(self, format, *args, **keywords):
        serializer = serializers.handlers[format]
        return serializer(self.root, *args, **keywords)



from __future__ import unicode_literals, print_function

import pytest

from treeoflife.file_storage import parse_line
from treeoflife.tracker import LoadError


class TestParseLine(object):
    def test_basic(self):
        line = u"\xfccategory: \xfcpersonal \xfcprojects"
        indent, is_metadata, node_id, node_type, text = parse_line(line)
        assert indent == 0
        assert not is_metadata
        assert node_id is None
        assert node_type == u"\xfccategory"
        assert text == u"\xfcpersonal \xfcprojects"

    def test_basic_nodeid(self):
        line = u"\xfccategory#asdfg: \xfcpersonal \xfcprojects"
        indent, is_metadata, node_id, node_type, text = parse_line(line)
        assert indent == 0
        assert not is_metadata
        assert node_id == u"asdfg"
        assert node_type == u"\xfccategory"
        assert text == u"\xfcpersonal \xfcprojects"

    def test_indent(self):
        line = u"    \xfcproject: \xfctodo \xfctracker"
        indent, is_metadata, node_id, node_type, text = parse_line(line)
        assert indent == 1
        assert not is_metadata
        assert node_id is None
        assert node_type == u"\xfcproject"
        assert text == u"\xfctodo \xfctracker"

    def test_indent_id(self):
        line = u"    \xfcproject#hjklo: \xfctodo \xfctracker"
        indent, is_metadata, node_id, node_type, text = parse_line(line)
        assert indent == 1
        assert not is_metadata
        assert node_id == u"hjklo"
        assert node_type == u"\xfcproject"
        assert text == u"\xfctodo \xfctracker"

    def test_empty_line(self):
        line = u""
        indent, is_metadata, node_id, node_type, text = parse_line(line)
        assert indent == 0
        assert not is_metadata
        assert node_id is None
        assert node_type == u""
        assert text is None

    def test_empty_line_indent(self):
        line = u"      "
        indent, is_metadata, node_id, node_type, text = parse_line(line)
        assert indent == 0
        assert not is_metadata
        assert node_id is None
        assert node_type == u""
        assert text is None

    def test_notext(self):
        line = u"        minor\xfc tasks"
        indent, is_metadata, node_id, node_type, text = parse_line(line)
        assert indent == 2
        assert not is_metadata
        assert node_id is None
        assert node_type == u"minor\xfc tasks"
        assert text is None

    def test_notext_id(self):
        line = u"        minor\xfc tasks#abcde"
        indent, is_metadata, node_id, node_type, text = parse_line(line)
        assert indent == 2
        assert not is_metadata
        assert node_id == u"abcde"
        assert node_type == u"minor\xfc tasks"
        assert text is None

    def test_metadata(self):
        line = u"    @option\xfc: \xfcvalue"
        indent, is_metadata, node_id, node_type, text = parse_line(line)
        assert indent == 1
        assert is_metadata
        assert node_id is None
        assert node_type == u"option\xfc"
        assert text == u"\xfcvalue"

    def test_metadata_notext(self):
        line = u"    @option\xfc"
        indent, is_metadata, node_id, node_type, text = parse_line(line)
        assert indent == 1
        assert is_metadata
        assert node_id is None
        assert node_type == u"option\xfc"
        assert text is None

    def test_bad_node_id(self):
        with pytest.raises(LoadError):
            parse_line(u"    option#longid: derp")
        with pytest.raises(LoadError):
            parse_line(u"    option#shrt: derp")

        with pytest.raises(LoadError):
            # invalid characters
            parse_line(u"    option#-----: derp")

    def test_metadata_node_id(self):
        with pytest.raises(LoadError):
            parse_line(u"    @option#badid: derp")

    def test_bad_indentation(self):
        with pytest.raises(LoadError):
            parse_line(u"  @option")

    def test_no_space(self):
        with pytest.raises(LoadError):
            parse_line(u"herp:derp")

    def test_non_unicode(self):
        with pytest.raises(AssertionError):
            parse_line(b"herp: derp")



from __future__ import unicode_literals, print_function

import inspect
import operator
import logging
import weakref
import random

from treeoflife.ordereddict import OrderedDict
from treeoflife.exceptions import (ListIntegrityError, LoadError,
        CantStartNodeError)
from treeoflife.util import HandlerDict
from treeoflife import file_storage

logger = logging.getLogger(__name__)


class _NodeCreatorTracker(HandlerDict):
    name = "creators"
    autodetect = False

    def create(self, node_type, text, parent, validate=True, nodeid=None):
        from treeoflife.nodes import import_all
        try:
            creator = self.creators[node_type]
        except KeyError:
            raise LoadError("No such node type: %r" % node_type)
        result = creator(node_type, text, parent, nodeid=nodeid)
        if result and validate:
            result._validate()
        return result

    def exists(self, node_type):
        return node_type in self.creators

    def values(self):
        return self.creators.values()

    def __call__(self, name):
        return self.add(name)


nodecreator = _NodeCreatorTracker()


class _NodeListIter(object):
    def __init__(self, nodelist, init_node=None, reverse=False):
        self.nodelist = nodelist

        if init_node is not None:
            self.node = init_node
        else:
            self.node = self.nodelist

        self.reverse = reverse

    def __iter__(self):
        return self

    def _prev(self):
        if self.node._prev_node is self.nodelist:
            raise StopIteration
        self.node = self.node._prev_node

        return self.node

    def _next(self):
        if self.node._next_node is self.nodelist:
            raise StopIteration
        self.node = self.node._next_node

        return self.node

    def next(self):
        if self.reverse:
            return self._prev()
        else:
            return self._next()

    def __repr__(self):
        return "NodeListIter(%r)" % (self.node,)


class _NodeListRoot(object):
    def __init__(self):
        self._next_node = self
        self._prev_node = self
        self.length = 0

    def insert(self, child, before=None, after=None):
        # note: this method makes a lot of assumptions about valid state!
        next_n = before
        prev_n = after

        if not next_n and not prev_n:
            next_n = self
            prev_n = self._prev_node

        elif next_n and not prev_n:
            prev_n = next_n._prev_node
        elif prev_n and not next_n:
            next_n = prev_n._next_node

        if prev_n._next_node is not next_n:
            raise ListIntegrityError(("prev_node %r has next neighbor %r, "
                    "but next_n is %r and has prev neighbor %r") %
                    (prev_n, prev_n._next_node, next_n, next_n._prev_node))
        if next_n._prev_node is not prev_n:
            raise ListIntegrityError(("next_node %r has prev neighbor %r, "
                    "but prev_n is %r and has next neighbor %r") %
                    (next_n, next_n._prev_node, prev_n, prev_n._next_node))

        next_n._prev_node = child
        child._next_node = next_n

        prev_n._next_node = child
        child._prev_node = prev_n
        self.length += 1

    def remove(self, child):
        before = child._prev_node
        after = child._next_node
#        child._prev_node = None
#        child._next_node = None
        before._next_node = after
        after._prev_node = before
        self.length -= 1

    @property
    def next_neighbor(self):
        if self._next_node is not self:
            return self._next_node
        return None

    @property
    def prev_neighbor(self):
        if self._prev_node is not self:
            return self._prev_node
        return None

    def __iter__(self):
        return _NodeListIter(self)

    def __reversed__(self):
        return _NodeListIter(self, reverse=True)

    def __len__(self):
        return self.length


class Node(object):
    multiline = False
    textless = False
    text_required = False
    toplevel = False
    can_activate = False
    options = ()
    children_of = None
    allowed_children = None
    preferred_parent = None

    #-------------------------------------------------#
    #                 initialization                  #
    #-------------------------------------------------#

    def __init__(self, node_type, text, parent=None, nodeid=None):
        self.id = nodeid

        self._init_children()

        self.__initing = True  # hack for assignment hooks

        self.root = None
        self.node_type = node_type
        self.parent = parent  # bootstrap parent for use in text hook
        self.text = text

        self.__initing = False
        self.parent = parent  # re-run the parent assignment hooks!

        if self.textless and text is not None:
            raise LoadError("%r node cannot have text" % self)

        if self.text_required and text is None:
            raise LoadError("%r node must have text" % self)

        if not self.multiline and text is not None and "\n" in text:
            raise LoadError("%r node cannot have newlines in text" % self)

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, newparent):
        self._parent = newparent
        if newparent is not None and self.root is not newparent.root:
            self.root = newparent.root

    @property
    def root(self):
        return self._root

    @root.setter
    def root(self, root):
        self._root = root
        if root is not None:
            if self.id is None:
                self.id = root.generate_id()
            if self.id in root.ids:  # pragma: no cover
                raise LoadError("Duplicate node IDs #%s: %r and %r" % (self.id,
                    self, self.root.ids[self.id]))
            root.ids[self.id] = self
        if (root is not None and not root.loading_in_progress
                and not self.__initing):
            self.load_finished()

    def _validate(self):
        if self.toplevel and self.parent != self.root:
            raise LoadError("%r node must be child of root node" % self)

        if self.children_of and self.parent.node_type not in self.children_of:
            raise LoadError("%s cannot be child of %s node" % (
                self._do_repr(parent=False),
                self.parent._do_repr(parent=False)
            ))

    def _init_children(self):
        self.children = _NodeListRoot()
        self.referred_to = set()

        self._next_node = None
        self._prev_node = None

    def continue_text(self, text):
        if not self.multiline:
            raise LoadError("%r node cannot have newline in text" % self)
        if self.text is None:
            raise LoadError("Cannot add new line to text of node %r,"
                    " since it has no text set" % self)

        self.text += "\n"
        self.text += text

    #-------------------------------------------------#
    #                     options                     #
    #-------------------------------------------------#

    def setoption(self, option, value):
        options = self._option_dict()
        try:
            handler = options[option]
        except KeyError:
            raise LoadError("node %r has no such option %r" % (self, option))

        handler.set(self, value)

    def _option_dict(self):
        try:
            return self._option_dict_cache
        except AttributeError:
            pass

        result = OrderedDict()
        mro = [self] + list(inspect.getmro(type(self)))
        seen = set()
        for resolve_element in reversed(mro):
            try:
                options = resolve_element.options
            except AttributeError:
                continue

            if id(options) in seen:
                continue
            seen.add(id(options))

            for option in options:
                result[option.name] = option
        self._option_dict_cache = result
        return result

    def option_values(self):
        result = []
        for name, handler in self._option_dict().items():
            show, item = handler.get(self)
            result.append((name, item, show))
        return result

    #-------------------------------------------------#
    #            iterators and movement               #
    #-------------------------------------------------#

    def iter_parents(self, skip_root=False):
        node = self

        while True:
            if skip_root and node is self.root:
                break
            if node is None:
                break
            yield node
            node = node.parent

    def iter_flat_children(self, skipper=False):
        stack = [iter(self.children)]

        def skip():
            skip.skip = True

        while len(stack):
            try:
                item = next(stack[-1])
            except StopIteration:
                stack.pop()
            else:
                if skipper:
                    skip.skip = False
                    yield len(stack), item, skip
                    if not skip.skip:
                        stack.append(iter(item.children))
                else:
                    yield len(stack), item
                    stack.append(iter(item.children))

    def iter_forward(self):
        return _NodeListIter(self.parent.children, init_node=self)

    def iter_backward(self):
        return _NodeListIter(self.parent.children, init_node=self,
                reverse=True)

    @property
    def next_neighbor(self):
        if self._next_node and self._next_node is self.parent.children:
            return None
        return self._next_node

    @property
    def prev_neighbor(self):
        if self._prev_node and self._prev_node is self.parent.children:
            return None
        return self._prev_node

    def find(self, query, rigid=False):
        from treeoflife import searching
        if rigid:
            query = searching.parse_single(query)
        else:
            query = searching.parse(query)
        return query(self)

    #-------------------------------------------------#
    #          adding and removing children           #
    #-------------------------------------------------#

    def create(self, *a, **kw):
        from treeoflife import searching
        creator = searching.parse_create(*a, **kw)
        return creator(self)

    def addchild(self, child, before=None, after=None):
        if (self.allowed_children is not None and
                child.node_type != "" and
                child.node_type not in self.allowed_children):
            raise LoadError("node %s cannot be child of %r" %
                    (child._do_repr(parent=False), self))
        if child.parent is not self and child.parent is not None:
            raise LoadError("node %r does not expect to be a child of %r" %
                    (child, self))
        if before and before.parent is not self:
            raise LoadError("node %r cannot be before node %r as child of %r" %
                    (child, before, self))
        if after and after.parent is not self:
            raise LoadError("node %r cannot be after node %r as child of %r" %
                    (child, after, self))
        self.children.insert(child, before, after)

        if child.parent is None:
            child.parent = self
            child._validate()

        ihook = getattr(child, "_insertion_hook", None)
        if ihook is not None:
            ihook()

        return child

    def createchild(self, node_type, text=None, nodeid=None, **keywords):
        node = self.root.nodecreator.create(node_type, text, self,
                nodeid=nodeid)
        return self.addchild(node, **keywords)

    def removechild(self, child):
        self.children.remove(child)
        child.parent = None

    def detach(self):
        if self.parent:
            self.parent.removechild(self)
        if self.root:
            del self.root.ids[self.id]
        return self

    def copy(self, parent=None, children=True, options=True, nodeid=None):
        if parent is None:
            parent = self.parent

        newnode = self.root.nodecreator.create(self.node_type, self.text,
                parent, nodeid=nodeid)
        if options:
            for option, value, show in self.option_values():
                if show:
                    newnode.setoption(option, value)
        if children:
            for child in self.children_export():
                child_copy = child.copy(parent=newnode)
                newnode.addchild(child_copy)
        return newnode

    #-------------------------------------------------#
    #        hooks for subclasses to override         #
    #-------------------------------------------------#

    @property
    def active_id(self):
        return self.id

    @classmethod
    def make_skeleton(self, root):
        pass

    def children_export(self):
        return list(self.children)

    def start(self):
        """
        Called to start the node
        (todo: isn't this a task-specific thing?)
        """
        if not self.can_activate:
            raise CantStartNodeError("can't start node %r" % self)

    def finish(self):
        """
        Called when the node is finished
        (todo: isn't this a task-specific thing?)
        """

    def load_finished(self):
        """
        Called when the  system is ready for the node to do
        things to the tree
        """

    def auto_add(self, creator, root):
        self.root = root
        if self.preferred_parent is not None:
            parent = creator.find(self.preferred_parent).one()
            parent.addchild(self)
            return parent
        else:
            return None

    def ui_dictify(self, result=None):
        """
        Called to create a json version of the node, for the ui
        """
        if result is None:
            result = {}

        if "options" not in result:
            options = [dict(zip(["type", "text"], option)) for option
                    in self.option_values() if option[-1]]
            if options:
                result["options"] = options
        if "children" not in result:
            children = [child.id for child in self.children]
            if children:
                result["children"] = children
        result["type"] = self.node_type
        result["text"] = self.text
        result["id"] = self.id
        return result

    def ui_graph(self):
        nodes = {self.id: self.ui_dictify()}
        for depth, node, skip in self.iter_flat_children(skipper=True):
            d = node.ui_dictify()
            if d is None:
                skip()
                continue
            nodes[node.id] = d
        return nodes

    def search_texts(self):
        """
        Called to determine what texts the node will match
        """
        return set([self.node_type]), set([self.text])

    def search_tags(self):
        """
        called to determine what tags the node will match
        """
        return set()

    def user_creation(self):
        """
        called when the node is created by user interaction
        (not called on deserialize)
        """

    #-------------------------------------------------#
    #                      misc                       #
    #-------------------------------------------------#

    def _search_tags(self):
        tags = set(self.search_tags())
        if self.can_activate:
            tags.add("can_activate")
        else:
            tags.add("cannot_activate")
        return tags

    def __str__(self):
        if self.text:
            result = "%s: %s" % (self.node_type, self.text)
        else:
            result = self.node_type
        return result.partition("\n")[0]

    def _format_repr_error(self, errors, name):
        import traceback
        result = "EXCEPTION IN REPR: %s\n" % name
        if errors is not None:
            errors.append(result + traceback.format_exc())
        return "[! " + name + " !]"

    def _format_repr_errors(self, errors):
        if not errors:
            return ""
        result = ""
        for error in errors:
            result += "\n\n"
            result += error
        return result

    def _do_repr(self, parent=True, exceptions=False):
        if exceptions:
            errors = []
        else:
            errors = None
        do_parent = parent
        del parent

        try:
            node_type = getattr(self, "node_type", None)
        except Exception as e:
            node_type = self._format_repr_error(
                    errors, "node_type")
        try:
            text = getattr(self, "text", None)
        except Exception as e:
            text = self._format_repr_error(
                    errors, "text")

        if do_parent:
            parent_repr = "None"
            try:
                parent = getattr(self, "parent", None)
            except Exception as e:
                parent = None
                parent_repr = self._format_repr_error(
                        errors, "self.parent")
            if parent is not None:
                try:
                    parent_repr = str(parent)
                except Exception as e:
                    parent_repr = self._format_repr_error(
                            errors, "str(parent)")

            return "<%s (%s) %r: %r>" % (
                    type(self).__name__,
                    parent_repr,
                    node_type,
                    text
            ) + self._format_repr_errors(errors)
        else:
            return "<%s %r: %r>" % (
                    type(self).__name__,
                    node_type,
                    text
            ) + self._format_repr_errors(errors)

    def __repr__(self):
        return self._do_repr()


@nodecreator('-')
def continue_text(node_type, text, parent, nodeid):
    parent.continue_text(text)


@nodecreator("")
class EmptyLineNode(Node):
    pass


class Option(object):
    incoming = None
    outgoing = None

    def __init__(self, name=None, incoming=None, outgoing=None):
        if name:
            self.name = name
        if incoming:
            self.incoming = incoming
        if outgoing:
            self.outgoing = outgoing
        assert self.name is not None, "name by arg or class-attr pls"

    def set(self, node, value):
        if self.incoming is not None:
            value = self.incoming(value)
        setattr(node, self.name, value)

    def get(self, node):
        try:
            value = getattr(node, self.name)
        except AttributeError:
            return False, None

        show = True
        if value is None:
            show = False

        if self.outgoing is not None and show:
            value = self.outgoing(value)

        return show, value


class BooleanOption(object):
    def __init__(self, name=None):
        if name:
            self.name = name
        assert self.name is not None, "name by arg or class-attr pls"

    def set(self, node, value):
        setattr(node, self.name, True)

    def get(self, node):
        value = getattr(node, self.name)
        show = bool(value)
        return show, None


class TreeRootNode(Node):
    def __init__(self, tracker, nodecreator, loading_in_progress=False):
        self.ids = weakref.WeakValueDictionary()
        self.nodecreator = nodecreator
        self.tracker = tracker
        self.loading_in_progress = loading_in_progress

        super(TreeRootNode, self).__init__("life", None, None, nodeid="00000")
        self.root = self

        self.editor_callback = None

        self.days = None
        self.active_node = None
        self.todo = None
        self.todo_review = None

    def load_finished(self):
        if getattr(self, "_alarm_hook", None) is not None:
            self._alarm_hook()

    def make_skeleton(self):
        for node_class in set(self.nodecreator.values()):
            func = getattr(node_class, "make_skeleton", None)
            if func:
                func(self)

    def activate(self, node, force=False):
        # jump to a particular node as active
        if not node.can_activate:
            if force:
                unfinish = getattr(node, "unfinish", None)
            else:
                unfinish = None
            if unfinish is None or not unfinish():
                logger.warn("Attempted to activate node: %r", node)
                return

        self.active_node = node
        for parent_node in list(node.iter_parents())[::-1]:
            try:
                parent_node.start()
            except CantStartNodeError:
                if parent_node is node:
                    raise

    def generate_id(self):
        for x in xrange(10):
            nodeid = "".join(random.choice(file_storage.nodeidchars)
                    for x in xrange(5))
            if nodeid in self.ids:
                continue
            return nodeid
        else:  # pragma: no cover
            raise Exception("10 tries to generate node id failed. wat? %s"
                    % nodeid)

#    def _skip_ignored(self, peergetter, node):
#        while node is not None:
#            if node.can_activate:
#                return node
#            node = peergetter(node)
#        return None
#
#    def _descend(self, peergetter, node):
#        while peergetter(node.children):
#            next_node = peergetter(node.children)
#            next_node = self._skip_ignored(peergetter, next_node)
#            if next_node:
#                node = next_node
#            else:
#                break
#        return node
#
#    def _navigate(self, peerattr):
#        #TODO: does not take skipping already-done ones into account
#        #TODO: it doesn't?
#        peergetter = operator.attrgetter(peerattr)
#
#        node = self.active_node
#
#        newnode = self._descend(peergetter, node)
#        if newnode is not node:
#            self.activate(newnode)
#            return
#
#        newnode = self._skip_ignored(peergetter, peergetter(node))
#        if newnode is not None:
#            self.activate(newnode)
#            node.finish()
#            return
#
#        newnode = self._skip_ignored(peergetter, node.parent)
#        if newnode:
#            self.activate(newnode)
#            node.finish()
#            return
#
#    def activate_next(self, *args, **keywords):
#        return self._navigate("next_neighbor", *args, **keywords)
#
#    def activate_prev(self, *args, **keywords):
#        return self._navigate("prev_neighbor", *args, **keywords)
#
#    def _create_related(self, node_type, text, relation, activate):
#        newnode = self.active_node.parent.createchild(node_type, text,
#                **{relation: self.active_node})
#        if activate:
#            self.activate(newnode)
#        return newnode
#
#    def create_before(self, node_type, text=None, activate=True):
#        return self._create_related(node_type, text, "before", activate)
#
#    def create_after(self, node_type, text=None, activate=True):
#        return self._create_related(node_type, text, "after", activate)
#
#    def create_child(self, node_type, text=None, activate=True):
#        newnode = self.active_node.createchild(node_type, text)
#        if activate:
#            self.activate(newnode)
#        return newnode
